import pandas as pd
import pysal
from pysal.explore.esda.moran import Moran_Local
from bokeh.plotting import figure
from bokeh.models import (GeoJSONDataSource, ColumnDataSource,
                          CategoricalColorMapper, Span,
                          HoverTool, Legend)
from bokeh.layouts import gridplot
from bokeh import palettes

from ._viz_utils import (bin_labels_choropleth, add_legend,
                         mask_local_auto, calc_data_aspect)

"""
Leightweight interactive visualizations in Bokeh.

TODO:
* We are not re-projection data into web-mercator atm,
    to allow plotting from raw coordinates.
    The user should however be aware of the projection of the used data.
"""

__author__ = ("Stefanie Lumnitz <stefanie.lumitz@gmail.com>")


def plot_choropleth(df, attribute, title=None, plot_width=500,
                    plot_height=500, method='quantiles',
                    k=5, reverse_colors=False, tools='', region_column=''):
    '''
    Plot Choropleth colored according to attribute

    Parameters
    ----------
    df : Geopandas dataframe
        Dataframe containign relevant shapes and attribute values.
    attribute : str
        Name of column containing attribute values of interest.
    title : str, optional
        Title of map. Default title=None
    plot_width : int, optional
        Width dimension of the figure in screen units/ pixels.
        Default = 500
    plot_height : int, optional
        Height dimension of the figure in screen units/ pixels.
        Default = 500
    method : str, optional
        Classification method to be used. Options supported:
        * 'quantiles' (default)
        * 'fisher-jenks'
        * 'equal-interval'
    k : int, optional
        Number of bins, assigning values to. Default k=5
    reverse_colors: boolean
        Reverses the color palette to show lightest colors for
        lowest values. Default reverse_colors=False
    tools : str, optional
        Tools used for bokeh plotting. Default = ''
    region_column : str, optional
        Column name containing region descpriptions/ names or polygone ids.
        Default = ''.

    Returns
    -------
    fig : Bokeh Figure instance
        Figure of Choropleth

    Examples
    --------
    >>> import pysal.lib.api as lp
    >>> from pysal.lib import examples
    >>> import geopandas as gpd
    >>> import pysal.explore.esda as esda
    >>> from pysal.viz.splot.bk import plot_choropleth
    >>> from bokeh.io import show

    >>> link = examples.get_path('columbus.shp')
    >>> df = gpd.read_file(link)
    >>> w = lp.Queen.from_dataframe(df)
    >>> w.transform = 'r'

    >>> TOOLS = "tap,help"
    >>> fig = plot_choropleth(df, 'HOVAL', title='columbus',
    ...                       reverse_colors=True, tools=TOOLS)
    >>> show(fig)
    '''
    # We're adding columns, do that on a copy rather than on the users' input
    df = df.copy()

    # Extract attribute values from df
    attribute_values = df[attribute].values

    # Create bin labels with bin_labels_choropleth()
    bin_labels = bin_labels_choropleth(df, attribute_values, method, k)

    # Initialize GeoJSONDataSource
    geo_source = GeoJSONDataSource(geojson=df.to_json())

    fig = _plot_choropleth_fig(geo_source, attribute, bin_labels,
                               bounds=df.total_bounds,
                               region_column=region_column, title=title,
                               plot_width=plot_width, plot_height=plot_height,
                               method=method, k=k,
                               reverse_colors=reverse_colors,
                               tools=tools)
    return fig


def _plot_choropleth_fig(geo_source, attribute, bin_labels, bounds,
                         region_column='', title=None,
                         plot_width=500, plot_height=500, method='quantiles',
                         k=5, reverse_colors=False, tools=''):
    colors = palettes.YlGnBu[k]
    if reverse_colors is True:
        colors.reverse()  # lightest color for lowest values

    # make data aspect ration match the figure aspect ratio
    # to avoid map distortion (1km=1km)
    x_min, x_max, y_min, y_max = calc_data_aspect(plot_height, plot_width,
                                                  bounds)

    # Create figure
    fig = figure(title=title, plot_width=plot_width, plot_height=plot_height,
                 tools=tools, x_range=(x_min, x_max), y_range=(y_min, y_max))
    # The use of `nonselection_fill_*` shouldn't be necessary,
    # but currently it is. This looks like a bug in Bokeh
    # where gridplot plus taptool chooses the underlay from the figure
    # that is clicked and applies it to the other figure as well.
    fill_color = {'field': 'labels_choro',
                  'transform': CategoricalColorMapper(palette=colors,
                                                      factors=bin_labels)}
    fig.patches('xs', 'ys', fill_alpha=0.7, fill_color=fill_color,
                line_color='white', nonselection_fill_alpha=0.2,
                nonselection_fill_color=fill_color,
                selection_line_color='firebrick',
                selection_fill_color=fill_color,
                line_width=0.5, source=geo_source)

    # add hover tool
    if 'hover' in tools:
        hover = fig.select_one(HoverTool)
        hover.point_policy = "follow_mouse"
        hover.tooltips = [("Region", "@" + region_column),
                          ("Attribute", "@" + attribute + "{0.0}"),
                          ]

    # add legend with add_legend()
    add_legend(fig, bin_labels, colors)

    # change layout
    fig.xgrid.grid_line_color = None
    fig.ygrid.grid_line_color = None
    fig.axis.visible = None
    return fig


def lisa_cluster(moran_loc, df, p=0.05, region_column='', title=None,
                 plot_width=500, plot_height=500, tools=''):
    '''
    Lisa Cluster map, coloured by local spatial autocorrelation

    Parameters
    ----------
    moran_loc : esda.moran.Moran_Local instance
        values of Moran's Local Autocorrelation Statistic
    df : geopandas dataframe instance
        In mask_local_auto(), assign df['labels'] per row. Note that
        ``df`` will be modified, so calling functions uses a copy of
        the user provided ``df``.
    p : float, optional
        The p-value threshold for significance. Points will
        be colored by significance.
    title : str, optional
        Title of map. Default title=None
    plot_width : int, optional
        Width dimension of the figure in screen units/ pixels.
        Default = 500
    plot_height : int, optional
        Height dimension of the figure in screen units/ pixels.
        Default = 500

    Returns
    -------
    fig : Bokeh figure instance
        Figure of LISA cluster map, colored by local spatial autocorrelation

    Examples
    --------
    >>> import pysal.lib.api as lp
    >>> from pysal.lib import examples
    >>> import geopandas as gpd
    >>> from pysal.explore.esda.moran import Moran_Local
    >>> from pysal.viz.splot.bk import lisa_cluster
    >>> from bokeh.io import show

    >>> link = examples.get_path('columbus.shp')
    >>> df = gpd.read_file(link)
    >>> y = df['HOVAL'].values
    >>> w = lp.Queen.from_dataframe(df)
    >>> w.transform = 'r'
    >>> moran_loc = Moran_Local(y, w)

    >>> TOOLS = "tap,reset,help"
    >>> fig = lisa_cluster(moran_loc, df, p=0.05, tools=TOOLS)
    >>> show(fig)
    '''
    # We're adding columns, do that on a copy rather than on the users' input
    df = df.copy()

    # add cluster_labels and colors5 in mask_local_auto
    cluster_labels, colors5, _, labels = mask_local_auto(moran_loc, p=0.05)
    df['labels_lisa'] = labels
    df['moranloc_psim'] = moran_loc.p_sim
    df['moranloc_q'] = moran_loc.q

    # load df into bokeh data source
    geo_source = GeoJSONDataSource(geojson=df.to_json())

    fig = _lisa_cluster_fig(geo_source, moran_loc, cluster_labels, colors5,
                            bounds=df.total_bounds,
                            region_column=region_column,
                            title=title, plot_width=plot_width,
                            plot_height=plot_height, tools=tools)
    return fig


def _lisa_cluster_fig(geo_source, moran_loc, cluster_labels, colors5,
                      bounds, region_column='', title=None, plot_width=500,
                      plot_height=500, tools=''):
    # make data aspect ration match the figure aspect ratio
    # to avoid map distortion (1km=1km)
    x_min, x_max, y_min, y_max = calc_data_aspect(plot_height, plot_width,
                                                  bounds)

    # Create figure
    fig = figure(title=title, toolbar_location='right',
                 plot_width=plot_width, plot_height=plot_height,
                 x_range=(x_min, x_max), y_range=(y_min, y_max), tools=tools)
    fill_color = {'field': 'labels_lisa',
                  'transform': CategoricalColorMapper(palette=colors5,
                                                      factors=cluster_labels)}
    fig.patches('xs', 'ys', fill_color=fill_color, fill_alpha=0.8,
                nonselection_fill_alpha=0.2,
                nonselection_fill_color=fill_color,
                line_color='white', selection_line_color='firebrick',
                selection_fill_color=fill_color,
                line_width=0.5, source=geo_source)

    if 'hover' in tools:
        # add hover tool
        hover = fig.select_one(HoverTool)
        hover.point_policy = "follow_mouse"
        hover.tooltips = [("Region", "@" + region_column),
                          ("Significance", "@moranloc_psim{0.00}"),
                          ("Quadrant", "@moranloc_q{0}")
                          ]

    # add legend with add_legend()
    add_legend(fig, cluster_labels, colors5)

    # change layout
    fig.xgrid.grid_line_color = None
    fig.ygrid.grid_line_color = None
    fig.axis.visible = None
    return fig


def moran_scatterplot(moran_loc, p=None, region_column='', plot_width=500,
          plot_height=500, tools=''):
    '''
    Moran Scatterplot, optional coloured by local spatial autocorrelation

    Parameters
    ----------
    moran_loc : esda.moran.Moran_Local instance
        values of Moran's Local Autocorrelation Statistic
    p : float, optional
        The p-value threshold for significance. Points will
        be colored by significance.
    plot_width : int, optional
        Width dimension of the figure in screen units/ pixels.
        Default = 500
    plot_height : int, optional
        Height dimension of the figure in screen units/ pixels.
        Default = 500

    Returns
    -------
    fig : Bokeh figure instance
        Figure of Moran Scatterplot, optionally colored by
        local spatial autocorrelation

    Examples
    --------
    >>> import pysal.lib.api as lp
    >>> from pysal.lib import examples
    >>> import geopandas as gpd
    >>> from pysal.explore.esda.moran import Moran_Local
    >>> from pysal.viz.splot.bk import moran_scatterplot
    >>> from bokeh.io import show

    >>> link = examples.get_path('columbus.shp')
    >>> df = gpd.read_file(link)
    >>> y = df['HOVAL'].values
    >>> w = lp.Queen.from_dataframe(df)
    >>> w.transform = 'r'
    >>> moran_loc = Moran_Local(y, w)

    >>> fig = moran_scatterplot(moran_loc, p=0.05)
    >>> show(fig)
    '''
    data = _moran_scatterplot_calc(moran_loc, p)
    source = ColumnDataSource(pd.DataFrame(data))
    fig = _moran_scatterplot_fig(source, p=p, region_column=region_column,
                     plot_width=plot_width, plot_height=plot_height,
                     tools=tools)
    return fig


def _moran_scatterplot_calc(moran_loc, p):
    lag = pysal.lib.weights.spatial_lag.lag_spatial(moran_loc.w, moran_loc.z)
    fit = pysal.model.spreg.OLS(moran_loc.z[:, None], lag[:, None])
    if p is not None:
        if not isinstance(moran_loc, Moran_Local):
            raise ValueError("`moran_loc` is not a esda.moran.Moran_Local instance")

        _, _, colors, _ = mask_local_auto(moran_loc, p=p)
    else:
        colors = 'black'

    data = {'moran_z': moran_loc.z, 'lag': lag,
            'colors': colors, 'fit_y': fit.predy.flatten(),
            'moranloc_psim': moran_loc.p_sim, 'moranloc_q': moran_loc.q}
    return data


def _moran_scatterplot_fig(source, p=None, title="Moran Scatterplot", region_column='',
               plot_width=500, plot_height=500, tools=''):
    """
    Parameters
    ----------
    source : Bokeh ColumnDatasource or GeoJSONDataSource instance
        The data source, should contain the columns ``moran_z`` and ``lag``,
        which will be used as x and y inputs of the scatterplot.
    """
    # Vertical line
    vline = Span(location=0, dimension='height', line_color='lightskyblue',
                 line_width=2, line_dash='dashed')
    # Horizontal line
    hline = Span(location=0, dimension='width', line_color='lightskyblue',
                 line_width=2, line_dash='dashed')

    # Create figure
    fig = figure(title=title, x_axis_label='Response',
                 y_axis_label='Spatial Lag', toolbar_location='left',
                 plot_width=plot_width, plot_height=plot_height, tools=tools)
    fig.scatter(x='moran_z', y='lag', source=source, color='colors',
                size=8, fill_alpha=.6, selection_fill_alpha=1,
                selection_line_color='firebrick', selection_fill_color='colors')
    fig.renderers.extend([vline, hline])
    fig.xgrid.grid_line_color = None
    fig.ygrid.grid_line_color = None
    fig.line(x='lag', y='fit_y', source=source, line_width=2)  # fit line

    if 'hover' in tools:
        hover = fig.select_one(HoverTool)
        hover.point_policy = "follow_mouse"
        hover.tooltips = [("Region", "@" + region_column),
                          ("Significance", "@moranloc_psim{0.00}"),
                          ("Quadrant", "@moranloc_q{0}")
                          ]
    return fig


def plot_local_autocorrelation(moran_loc, df, attribute, p=0.05,
                               region_column='', plot_width=350,
                               plot_height=400, method='quantiles', k=5,
                               reverse_colors=False):
    """
    Plot Moran Scatterplot, LISA cluster and Choropleth
    for Local Spatial Autocorrelation Analysis

    Parameters
    ----------
    moran_loc : esda.moran.Moran_Local instance
        values of Moran's Local Autocorrelation Statistic
    df : Geopandas dataframe
        Dataframe containing relevant polygon and attribute values.
    attribute : str
        Name of column containing attribute values of interest.
    plot_width : int, optional
        Width dimension of the figure in screen units/ pixels.
        Default = 250
    plot_height : int, optional
        Height dimension of the figure in screen units/ pixels.
        Default = 300
    method : str, optional
        Classification method to be used. Options supported:
        * 'quantiles' (default)
        * 'fisher-jenks'
        * 'equal-interval'
    k : int, optional
        Number of bins, assigning values to. Default k=5
    reverse_colors: boolean
        Reverses the color palette to show lightest colors for
        lowest values in Choropleth map. Default reverse_colors=False

    Returns
    -------
    fig : Bokeh Figure instance
        Figure of Choropleth

    Examples
    --------
    >>> import pysal.lib.api as lp
    >>> from pysal.lib import examples
    >>> import geopandas as gpd
    >>> from pysal.explore.esda.moran import Moran_Local
    >>> from pysal.viz.splot.bk import plot_local_autocorrelation
    >>> from bokeh.io import show

    >>> link = examples.get_path('columbus.shp')
    >>> df = gpd.read_file(link)
    >>> y = df['HOVAL'].values
    >>> w = lp.Queen.from_dataframe(df)
    >>> w.transform = 'r'
    >>> moran_loc = Moran_Local(y, w)

    >>> fig = plot_local_autocorrelation(moran_loc, df, 'HOVAL',
                                         reverse_colors=True)
    >>> show(fig)
    """
    # We're adding columns, do that on a copy rather than on the users' input
    df = df.copy()

    # Add relevant results for moran_scatterplot as columns to geodataframe
    moran_scatterplot_data = _moran_scatterplot_calc(moran_loc, p)
    for key in moran_scatterplot_data:
        df[key] = moran_scatterplot_data[key]

    # add cluster_labels and colors5 in mask_local_auto
    cluster_labels, colors5, _, labels = mask_local_auto(moran_loc, p=0.05)
    df['labels_lisa'] = labels
    df['moranloc_psim'] = moran_loc.p_sim
    df['moranloc_q'] = moran_loc.q
    # Extract attribute values from df
    attribute_values = df[attribute].values
    # Create bin labels with bin_labels_choropleth()
    bin_labels = bin_labels_choropleth(df, attribute_values, method, k)

    # load df into bokeh data source
    geo_source = GeoJSONDataSource(geojson=df.to_json())

    TOOLS = "tap,reset,help,hover"

    scatter = _moran_scatterplot_fig(geo_source, p=p, region_column=region_column,
                         title="Local Spatial Autocorrelation",
                         plot_width=int(plot_width*1.15),
                         plot_height=plot_height,
                         tools=TOOLS)
    LISA = _lisa_cluster_fig(geo_source, moran_loc, cluster_labels, colors5,
                             bounds=df.total_bounds,
                             region_column=region_column,
                             plot_width=plot_width,
                             plot_height=plot_height, tools=TOOLS)
    choro = _plot_choropleth_fig(geo_source, attribute, bin_labels,
                                 bounds=df.total_bounds,
                                 region_column=region_column,
                                 reverse_colors=reverse_colors,
                                 plot_width=plot_width,
                                 plot_height=plot_height,
                                 tools=TOOLS)

    fig = gridplot([[scatter, LISA, choro]],
                   sizing_mode='scale_width')
    return fig



import numpy as np
from pysal.model.spglm.family import Gaussian, Binomial, Poisson
from pysal.model.spglm.glm import GLM
from .diagnostics import get_AICc

def summaryModel(self):
    summary = '=' * 75 + '\n'
    summary += "%-54s %20s\n" % ('Model type', self.family.__class__.__name__)
    summary += "%-60s %14d\n" % ('Number of observations:', self.n)
    summary += "%-60s %14d\n\n" % ('Number of covariates:', self.k)
    return summary

def summaryGLM(self):
    
    XNames = ["X"+str(i) for i in range(self.k)]
    glm_rslt = GLM(self.model.y,self.model.X,constant=False,family=self.family).fit()

    summary = "%s\n" %('Global Regression Results')
    summary += '-' * 75 + '\n'
    
    if isinstance(self.family, Gaussian):
        summary += "%-62s %12.3f\n" %  ('Residual sum of squares:', glm_rslt.deviance)
        summary += "%-62s %12.3f\n" %  ('Log-likelihood:', glm_rslt.llf)
        summary += "%-62s %12.3f\n" %  ('AIC:', glm_rslt.aic)
        summary += "%-62s %12.3f\n" %  ('AICc:', get_AICc(glm_rslt))
        summary += "%-62s %12.3f\n" %  ('BIC:', glm_rslt.bic)
        summary += "%-62s %12.3f\n" %  ('R2:', glm_rslt.D2)
        summary += "%-62s %12.3f\n\n" % ('Adj. R2:', glm_rslt.adj_D2)
    else:
        summary += "%-62s %12.3f\n" %  ('Deviance:', glm_rslt.deviance)
        summary += "%-62s %12.3f\n" %  ('Log-likelihood:', glm_rslt.llf)
        summary += "%-62s %12.3f\n" %  ('AIC:', glm_rslt.aic)
        summary += "%-62s %12.3f\n" %  ('AICc:', get_AICc(glm_rslt))
        summary += "%-62s %12.3f\n" %  ('BIC:', glm_rslt.bic)
        summary += "%-62s %12.3f\n" %  ('Percent deviance explained:', glm_rslt.D2)
        summary += "%-62s %12.3f\n\n" % ('Adj. percent deviance explained:', glm_rslt.adj_D2)
    
    summary += "%-31s %10s %10s %10s %10s\n" % ('Variable', 'Est.', 'SE' ,'t(Est/SE)', 'p-value')
    summary += "%-31s %10s %10s %10s %10s\n" % ('-'*31, '-'*10 ,'-'*10, '-'*10,'-'*10)
    for i in range(self.k):
        summary += "%-31s %10.3f %10.3f %10.3f %10.3f\n" % (XNames[i], glm_rslt.params[i], glm_rslt.bse[i], glm_rslt.tvalues[i], glm_rslt.pvalues[i])
    summary += "\n"
    return summary

def summaryGWR(self):
    XNames = ["X"+str(i) for i in range(self.k)]
    
    summary = "%s\n" %('Geographically Weighted Regression (GWR) Results')
    summary += '-' * 75 + '\n'

    if self.model.fixed:
        summary += "%-50s %20s\n" % ('Spatial kernel:', 'Fixed ' + self.model.kernel)
    else:
        summary += "%-54s %20s\n" % ('Spatial kernel:', 'Adaptive ' + self.model.kernel)

    summary += "%-62s %12.3f\n" % ('Bandwidth used:', self.model.bw)

    summary += "\n%s\n" % ('Diagnostic information')
    summary += '-' * 75 + '\n'
    
    if isinstance(self.family, Gaussian):
        
        summary += "%-62s %12.3f\n" % ('Residual sum of squares:', self.resid_ss)
        summary += "%-62s %12.3f\n" % ('Effective number of parameters (trace(S)):', self.tr_S)
        summary += "%-62s %12.3f\n" % ('Degree of freedom (n - trace(S)):', self.df_model)
        summary += "%-62s %12.3f\n" % ('Sigma estimate:', np.sqrt(self.sigma2))
        summary += "%-62s %12.3f\n" % ('Log-likelihood:', self.llf)
        summary += "%-62s %12.3f\n" % ('AIC:', self.aic)
        summary += "%-62s %12.3f\n" % ('AICc:', self.aicc)
        summary += "%-62s %12.3f\n" % ('BIC:', self.bic)
        summary += "%-62s %12.3f\n" % ('R2:', self.R2)
    else:
        summary += "%-62s %12.3f\n" % ('Effective number of parameters (trace(S)):', self.tr_S)
        summary += "%-62s %12.3f\n" % ('Degree of freedom (n - trace(S)):', self.df_model)
        summary += "%-62s %12.3f\n" % ('Log-likelihood:', self.llf)
        summary += "%-62s %12.3f\n" % ('AIC:', self.aic)
        summary += "%-62s %12.3f\n" % ('AICc:', self.aicc)
        summary += "%-62s %12.3f\n" % ('BIC:', self.bic)
        #summary += "%-60s %12.6f\n" % ('Percent deviance explained:', 0)


    summary += "%-62s %12.3f\n" % ('Adj. alpha (95%):', self.adj_alpha[1])
    summary += "%-62s %12.3f\n" % ('Adj. critical t value (95%):', self.critical_tval(self.adj_alpha[1]))

    summary += "\n%s\n" % ('Summary Statistics For GWR Parameter Estimates')
    summary += '-' * 75 + '\n'
    summary += "%-20s %10s %10s %10s %10s %10s\n" % ('Variable', 'Mean' ,'STD', 'Min' ,'Median', 'Max')
    summary += "%-20s %10s %10s %10s %10s %10s\n" % ('-'*20, '-'*10 ,'-'*10, '-'*10 ,'-'*10, '-'*10)
    for i in range(self.k):
        summary += "%-20s %10.3f %10.3f %10.3f %10.3f %10.3f\n" % (XNames[i], np.mean(self.params[:,i]) ,np.std(self.params[:,i]),np.min(self.params[:,i]) ,np.median(self.params[:,i]), np.max(self.params[:,i]))

    summary += '=' * 75 + '\n'

    return summary



def summaryMGWR(self):
    
    XNames = ["X"+str(i) for i in range(self.k)]
    
    summary = ''
    summary += "%s\n" %('Multi-Scale Geographically Weighted Regression (MGWR) Results')
    summary += '-' * 75 + '\n'
    
    if self.model.fixed:
        summary += "%-50s %20s\n" % ('Spatial kernel:', 'Fixed ' + self.model.kernel)
    else:
        summary += "%-54s %20s\n" % ('Spatial kernel:', 'Adaptive ' + self.model.kernel)

    summary += "%-54s %20s\n" % ('Criterion for optimal bandwidth:', self.model.selector.criterion)

    if self.model.selector.rss_score:
        summary += "%-54s %20s\n" % ('Score of Change (SOC) type:', 'RSS')
    else:
        summary += "%-54s %20s\n" % ('Score of Change (SOC) type:', 'Smoothing f')

    summary += "%-54s %20s\n\n" % ('Termination criterion for MGWR:', self.model.selector.tol_multi)

    summary += "%s\n" %('MGWR bandwidths')
    summary += '-' * 75 + '\n'
    summary += "%-15s %14s %10s %16s %16s\n" % ('Variable', 'Bandwidth', 'ENP_j','Adj t-val(95%)','Adj alpha(95%)')
    for j in range(self.k):
        summary += "%-14s %15.3f %10.3f %16.3f %16.3f\n" % (XNames[j], self.model.bw[j], self.ENP_j[j],self.critical_tval()[j],self.adj_alpha_j[j,1])

    summary += "\n%s\n" % ('Diagnostic information')
    summary += '-' * 75 + '\n'
    
    summary += "%-62s %12.3f\n" % ('Residual sum of squares:', self.resid_ss)
    summary += "%-62s %12.3f\n" % ('Effective number of parameters (trace(S)):', self.tr_S)
    summary += "%-62s %12.3f\n" % ('Degree of freedom (n - trace(S)):', self.df_model)
    
    summary += "%-62s %12.3f\n" % ('Sigma estimate:', np.sqrt(self.sigma2))
    summary += "%-62s %12.3f\n" % ('Log-likelihood:', self.llf)
    summary += "%-62s %12.3f\n" % ('AIC:', self.aic)
    summary += "%-62s %12.3f\n" % ('AICc:', self.aicc)
    summary += "%-62s %12.3f\n" % ('BIC:', self.bic)

    summary += "\n%s\n" % ('Summary Statistics For MGWR Parameter Estimates')
    summary += '-' * 75 + '\n'
    summary += "%-20s %10s %10s %10s %10s %10s\n" % ('Variable', 'Mean' ,'STD', 'Min' ,'Median', 'Max')
    summary += "%-20s %10s %10s %10s %10s %10s\n" % ('-'*20, '-'*10 ,'-'*10, '-'*10 ,'-'*10, '-'*10)
    for i in range(self.k):
        summary += "%-20s %10.3f %10.3f %10.3f %10.3f %10.3f\n" % (XNames[i], np.mean(self.params[:,i]) ,np.std(self.params[:,i]),np.min(self.params[:,i]) ,np.median(self.params[:,i]), np.max(self.params[:,i]))

    summary += '=' * 75 + '\n'
    return summary






"""
``splot.esda``
===============

Provides visualisations for the `esda` subpackage.
`esda` provides tools for exploratory spatial data analysis that consider the role of space in a distribution of attribute values.

Moran analytics
---------------

.. autosummary::
   :toctree: generated/
   
   moran_scatterplot
   plot_moran_simulation
   plot_moran
   plot_moran_bv_simulation
   plot_moran_bv
   lisa_cluster
   plot_local_autocorrelation
   moran_facet

"""

from ._viz_esda_mpl import (moran_scatterplot,
                            plot_moran_simulation,
                            plot_moran,
                            plot_moran_bv_simulation,
                            plot_moran_bv,
                            lisa_cluster,
                            plot_local_autocorrelation,
                            moran_facet)



__version__ = "1.1.0"
"""
:mod:`spaghetti` --- Spatial Graphs: Networks, Topology, & Inference
====================================================================

"""
from .network import Network, PointPattern, SimulatedPointPattern, SortedEdges
from .analysis import NetworkBase, NetworkG, NetworkK, NetworkF
from .analysis import gfunction, kfunction, ffunction
from .util import compute_length, get_neighbor_distances, generatetree
from .util import dijkstra, dijkstra_mp
from .util import squared_distance_point_segment, snap_points_on_segments



__version__ = "2.0.1"
"""
:mod:`pysal.explore.esda` --- Exploratory Spatial Data Analysis
=================================================

"""
from .moran import (Moran, Moran_BV, Moran_BV_matrix,
                    Moran_Local, Moran_Local_BV, 
                    Moran_Rate, Moran_Local_Rate)
from .getisord import G, G_Local
from .geary import Geary
from .join_counts import Join_Counts
from .gamma import Gamma



# GWR kernel function specifications

__author__ = "Taylor Oshan tayoshan@gmail.com"

import scipy
from scipy.spatial.kdtree import KDTree
import numpy as np
from scipy.spatial.distance import cdist as cdist_scipy
from math import radians, sin, cos, sqrt, asin

#adaptive specifications should be parameterized with nn-1 to match original gwr
#implementation. That is, pysal counts self neighbors with knn automatically.

def fix_gauss(coords, bw, points=None, dmat=None,sorted_dmat=None,spherical=False):
    """
    Fixed Gaussian kernel.
    """
    w = _Kernel(coords, function='gwr_gaussian', bandwidth=bw,
            truncate=False, points=points, dmat=dmat,sorted_dmat=sorted_dmat,spherical=spherical)
    return w.kernel

def adapt_gauss(coords, nn, points=None, dmat=None,sorted_dmat=None,spherical=False):
    """
    Spatially adaptive Gaussian kernel.
    """
    w = _Kernel(coords, fixed=False, k=nn-1, function='gwr_gaussian',
            truncate=False, points=points, dmat=dmat,sorted_dmat=sorted_dmat,spherical=spherical)
    return w.kernel

def fix_bisquare(coords, bw, points=None, dmat=None,sorted_dmat=None,spherical=False):
    """
    Fixed bisquare kernel.
    """
    w = _Kernel(coords, function='bisquare', bandwidth=bw, points=points, dmat=dmat,sorted_dmat=sorted_dmat,spherical=spherical)
    return w.kernel

def adapt_bisquare(coords, nn, points=None, dmat=None,sorted_dmat=None,spherical=False):
    """
    Spatially adaptive bisquare kernel.
    """
    w = _Kernel(coords, fixed=False, k=nn-1, function='bisquare', points=points, dmat=dmat,sorted_dmat=sorted_dmat,spherical=spherical)
    return w.kernel

def fix_exp(coords, bw, points=None, dmat=None,sorted_dmat=None,spherical=False):
    """
    Fixed exponential kernel.
    """
    w = _Kernel(coords, function='exponential', bandwidth=bw,
            truncate=False, points=points, dmat=dmat,sorted_dmat=sorted_dmat,spherical=spherical)
    return w.kernel

def adapt_exp(coords, nn, points=None, dmat=None,sorted_dmat=None,spherical=False):
    """
    Spatially adaptive exponential kernel.
    """
    w = _Kernel(coords, fixed=False, k=nn-1, function='exponential',
            truncate=False, points=points, dmat=dmat,sorted_dmat=sorted_dmat,spherical=spherical)
    return w.kernel

from scipy.spatial.distance import cdist

#Customized Kernel class user for GWR because the default PySAL kernel class
#favors memory optimization over speed optimizations and GWR often needs the 
#speed optimization since it is not always assume points far awary
# are truncated #to zero

def cdist(coords1,coords2,spherical):
    def _haversine(lon1, lat1, lon2, lat2):
        R = 6371.0 # Earth radius in kilometers
        dLat = radians(lat2 - lat1)
        dLon = radians(lon2 - lon1)
        lat1 = radians(lat1)
        lat2 = radians(lat2)
        a = sin(dLat/2)**2 + cos(lat1)*cos(lat2)*sin(dLon/2)**2
        c = 2*asin(sqrt(a))
        return R * c
    n = len(coords1)
    m = len(coords2)
    dmat = np.zeros((n,n))

    if spherical:
        for i in range(n) :
            for j in range(m):
                dmat[i,j] = _haversine(coords1[i][0], coords1[i][1], coords2[j][0], coords2[j][1])
    else:
        dmat = cdist_scipy(coords1,coords2)

    return dmat

class _Kernel(object):
    """
    GWR kernel function specifications.

    """
    def __init__(self, data, bandwidth=None, fixed=True, k=None,
                 function='triangular', eps=1.0000001, ids=None, truncate=True,
                 points=None, dmat=None,sorted_dmat=None, spherical=False): #Added truncate flag
        

        if issubclass(type(data), scipy.spatial.KDTree):
            self.data = data.data
            data = self.data
        else:
            self.data = data
        if k is not None:
            self.k = int(k) + 1
        else:
            self.k = k
        
        self.spherical = spherical
        self.searching = True
        
        if dmat is None:
            self.searching = False
        
        if self.searching:
            self.dmat = dmat
            self.sorted_dmat = sorted_dmat
        else:
            if points is None:
                self.dmat = cdist(self.data, self.data, self.spherical)
            else:
                self.points = points
                self.dmat = cdist(self.points, self.data, self.spherical)

        self.function = function.lower()
        self.fixed = fixed
        self.eps = eps
        self.trunc = truncate
        if bandwidth:
            try:
                bandwidth = np.array(bandwidth)
                bandwidth.shape = (len(bandwidth), 1)
            except:
                bandwidth = np.ones((len(data), 1), 'float') * bandwidth
            self.bandwidth = bandwidth
        else:
            self._set_bw()
        self.kernel = self._kernel_funcs(self.dmat/self.bandwidth)
        if self.trunc:
            mask = np.repeat(self.bandwidth, len(self.data), axis=1)
            self.kernel[(self.dmat >= mask)] = 0
                
    def _set_bw(self):
        if self.searching:
            if self.k is not None:
                dmat = self.sorted_dmat[:,:self.k]
            else:
                dmat = self.dmat
        else:
            if self.k is not None:
                dmat = np.sort(self.dmat)[:,:self.k]
            else:
                dmat = self.dmat
        
        if self.fixed:
            # use max knn distance as bandwidth
            bandwidth = dmat.max() * self.eps
            n = len(self.data)
            self.bandwidth = np.ones((n, 1), 'float') * bandwidth
        else:
            # use local max knn distance
            self.bandwidth = dmat.max(axis=1) * self.eps
            self.bandwidth.shape = (self.bandwidth.size, 1)

    def _kernel_funcs(self, zs):
        # functions follow Anselin and Rey (2010) table 5.4
        if self.function == 'triangular':
            return 1 - zs
        elif self.function == 'uniform':
            return np.ones(zi.shape) * 0.5
        elif self.function == 'quadratic':
            return (3. / 4) * (1 - zs ** 2)
        elif self.function == 'quartic':
            return (15. / 16) * (1 - zs ** 2) ** 2
        elif self.function == 'gaussian':
            c = np.pi * 2
            c = c ** (-0.5)
            return c * np.exp(-(zs ** 2) / 2.)
        elif self.function == 'gwr_gaussian':
            return np.exp(-0.5*(zs)**2)
        elif self.function == 'bisquare':
            return (1-(zs)**2)**2
        elif self.function =='exponential':
            return np.exp(-zs)
        else:
            print('Unsupported kernel function', self.function)



from collections import defaultdict, OrderedDict
import os
import pickle
import copy
import numpy as np
from .analysis import NetworkG, NetworkK, NetworkF
from . import util
from pysal.lib import cg, examples, weights
try:
    from pysal.lib import open
except ImportError:
    import pysal.lib
    open = pysal.lib.io.open

__all__ = ["Network", "PointPattern", "NetworkG", "NetworkK", "NetworkF"]


class Network:
    """Spatially-constrained network representation
    and analytical functionality.
    
    Parameters
    ----------
    
    in_data : geopandas.GeoDataFrame or str
        The input geographic data. Either (1) a path to a shapefile (str);
        or (2) a geopandas.GeoDataFrame.
    
    node_sig : int
        Round the x and y coordinates of all nodes to node_sig significant
        digits (combined significant digits on the left and right of the
        decimal place). Default is 11. Set to None for no rounding.
    
    unique_segs : bool
        If True (default), keep only unique segments (i.e., prune out any
        duplicated segments). If False keep all segments.
    
    extractgraph : bool
        If True, extract a graph-theoretic object with no degree 2 nodes.
        Default is True.
    
    Attributes
    ----------
    
    in_data : str
        The input shapefile. This must be in .shp format.
    
    adjacencylist : list
        List of lists storing node adjacency.
    
    nodes : dict
        Keys are tuples of node coords and values are the node ID.
    
    edge_lengths : dict
        Keys are tuples of sorted node IDs representing an edge and values
        are the length.
    
    pointpatterns : dict
        Keys are a string name of the pattern and values are point pattern
        class instances.
    
    node_coords : dict
        Keys are the node ID and values are the (x,y) coordinates
        inverse to nodes.
    
    edges : list
        List of edges, where each edge is a sorted tuple of node IDs.
    
    node_list : list
        List of node IDs.
    
    alldistances : dict
        Keys are the node IDs (int). Values are tuples with two elements as
        follows (1) a list of the shortest path distances; (2) a dict with the
        key being the id of the destination node and the value being a list of
        the shortest path.
    
    distancematrix : numpy.ndarray
        all network nodes (non-observations) distance matrix.
    
    graphedges : list
        tuples of graph edge ids.
    
    graph_lengths : dict
        Keys are the graph edge ids (tuple). Values are the graph edge lenght
        (float).
    
    Examples
    --------
    
    Instantiate an instance of a network.
    
    >>> import pysal.explore.spaghetti as spgh
    >>> streets_file = examples.get_path('streets.shp')
    >>> ntw = spgh.Network(in_data=streets_file)
    
    Snap point observations to the network with attribute information.
    
    >>> crimes_file = examples.get_path('crimes.shp')
    >>> ntw.snapobservations(crimes_file, 'crimes', attribute=True)
   
    And without attribute information.
   
    >>> schools_file = examples.get_path('schools.shp')
    >>> ntw.snapobservations(schools_file, 'schools', attribute=False)
    
    """
    
    def __init__(self, in_data=None, node_sig=11,
                 unique_segs=True, extractgraph=True):
        """
        """
        if in_data is not None:
            self.in_data = in_data
            self.node_sig = node_sig
            self.unique_segs = unique_segs
            
            self.adjacencylist = defaultdict(list)
            self.nodes = {}
            self.edge_lengths = {}
            self.edges = []
            
            self.pointpatterns = {}
            
            self._extractnetwork()
            self.node_coords = dict((v, k) for k, v in self.nodes.items())
            
            # This is a spatial representation of the network.
            self.edges = sorted(self.edges)
            
            # Extract the graph.
            if extractgraph:
                self.extractgraph()
                
            self.node_list = sorted(self.nodes.values())


    def _round_sig(self, v):
        """Used internally to round the vertex to a set number of significant
        digits. If sig is set to 4, then the following are some possible
        results for a coordinate are as follows.
            (1) 0.0xxxx, (2) 0.xxxx, (3) x.xxx, (4) xx.xx,
            (5) xxx.x, (6) xxxx.0, (7) xxxx0.0
        """
        sig = self.node_sig
        if sig is None:
            return v
        out_v = [val if 0
                 else round(val, -int(np.floor(np.log10(np.fabs(val)))) +
                            (sig - 1))
                 for val in v]
        
        return tuple(out_v)


    def _extractnetwork(self):
        """Used internally to extract a network from a polyline shapefile.
        """
        nodecount = 0
        if isinstance(self.in_data, str):
            shps = open(self.in_data)
        else:
            shps = self.in_data.geometry
        for shp in shps:
            vertices = weights._contW_lists._get_verts(shp)
            for i, v in enumerate(vertices[:-1]):
                v = self._round_sig(v)
                try:
                    vid = self.nodes[v]
                except KeyError:
                    self.nodes[v] = vid = nodecount
                    nodecount += 1
                v2 = self._round_sig(vertices[i + 1])
                try:
                    nvid = self.nodes[v2]
                except KeyError:
                    self.nodes[v2] = nvid = nodecount
                    nodecount += 1
                    
                self.adjacencylist[vid].append(nvid)
                self.adjacencylist[nvid].append(vid)
                
                # Sort the edges so that mono-directional keys can be stored.
                edgenodes = sorted([vid, nvid])
                edge = tuple(edgenodes)
                self.edges.append(edge)
                length = util.compute_length(v, vertices[i + 1])
                self.edge_lengths[edge] = length
        if self.unique_segs:
            # Remove duplicate edges and duplicate adjacent nodes.
            self.edges = list(set(self.edges))
            for k, v in self.adjacencylist.items():
                self.adjacencylist[k] = list(set(v))


    def extractgraph(self):
        """Using the existing network representation, create a graph-theoretic
        representation by removing all nodes with a neighbor incidence of two 
        (non-articulation points). That is, we assume these nodes are bridges
        between nodes with higher incidence.
        """
        self.graphedges = []
        self.graph_lengths = {}
        
        # Find all nodes with cardinality 2. (non-articulation points)
        segment_nodes = []
        for k, v in self.adjacencylist.items():
            # len(v) == 1 #cul-de-sac
            # len(v) == 2 #bridge segment
            # len(v) > 2 #intersection
            if len(v) == 2:
                segment_nodes.append(k)
                
        # Start with a copy of the spatial representation and
        # iteratively remove edges deemed to be segments.
        self.graphedges = copy.deepcopy(self.edges)
        self.graph_lengths = copy.deepcopy(self.edge_lengths)
        # Mapping all the 'network edges' contained within a single
        # 'graph represented' edge.
        self.graph_to_edges = {}
        
        # build up bridges "rooted" on the initial non-articulation points
        bridges = []
        for s in segment_nodes:
            bridge = [s]
            neighbors = self._yieldneighbor(s, segment_nodes, bridge)
            while neighbors:
                cnode = neighbors.pop()
                segment_nodes.remove(cnode)
                bridge.append(cnode)
                newneighbors = self._yieldneighbor(cnode,
                                                   segment_nodes,
                                                   bridge)
                neighbors += newneighbors
            bridges.append(bridge)
            
        for bridge in bridges:
            if len(bridge) == 1:
                n = self.adjacencylist[bridge[0]]
                newedge = tuple(sorted([n[0], n[1]]))
                # Identify the edges to be removed.
                e1 = tuple(sorted([bridge[0], n[0]]))
                e2 = tuple(sorted([bridge[0], n[1]]))
                # Remove them from the graph.
                self.graphedges.remove(e1)
                self.graphedges.remove(e2)
                # Remove from the edge lengths.
                length_e1 = self.edge_lengths[e1]
                length_e2 = self.edge_lengths[e2]
                self.graph_lengths.pop(e1, None)
                self.graph_lengths.pop(e2, None)
                self.graph_lengths[newedge] = length_e1 + length_e2
                # Update the pointers.
                self.graph_to_edges[e1] = newedge
                self.graph_to_edges[e2] = newedge
            else:
                cumulative_length = 0
                startend = {}
                redundant = set([])
                for b in bridge:
                    for n in self.adjacencylist[b]:
                        if n not in bridge:
                            startend[b] = n
                        else:
                            redundant.add(tuple(sorted([b, n])))
                            
                newedge = tuple(sorted(startend.values()))
                for k, v in startend.items():
                    redundant.add(tuple(sorted([k, v])))
                    
                for r in redundant:
                    self.graphedges.remove(r)
                    cumulative_length += self.edge_lengths[r]
                    self.graph_lengths.pop(r, None)
                    self.graph_to_edges[r] = newedge
                self.graph_lengths[newedge] = cumulative_length
                
            self.graphedges.append(newedge)
        self.graphedges = sorted(self.graphedges)


    def _yieldneighbor(self, node, segment_nodes, bridge):
        """Used internally, this method traverses a bridge segement to find
        the source and destination nodes.
        
        Parameters
        ----------
        
        node : int
            node id
        
        segment_nodes : list
            all non-articulation points in the network (degree-2 nodes).
        
        bridge : list
            inital bridge list containing only `node`
        
        Returns
        -------
        
        n : list
            nodes to keep (articulation points)
        
        """
        n = []
        # get all nodes adjacent to `node`
        for i in self.adjacencylist[node]:
            if i in segment_nodes and i not in bridge:
                n.append(i)
                
        return n


    def contiguityweights(self, graph=True, weightings=None):
        """Create a contiguity based W object.
        
        Parameters
        ----------
        
        graph : bool
            {True, False} controls whether the W is generated using the spatial
            representation or the graph representation.
        
        weightings : dict
            Dict of lists of weightings for each edge.
        
        Returns
        -------
        
         W : pysal.lib.weights.weights.W
            A PySAL W Object representing the binary adjacency of the network.
        
        Examples
        --------
        
        Instantiate an instance of a network.
        
        >>> import pysal.explore.spaghetti as spgh
        >>> from pysal.lib import examples
        >>> import pysal.explore.esda
        >>> import numpy as np
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        
        Snap point observations to the network with attribute information.
        
        >>> ntw.snapobservations(examples.get_path('crimes.shp'),
        ...                      'crimes', attribute=True)
        
        Find counts per netowrk edge.
        
        >>> counts = ntw.count_per_edge(ntw.pointpatterns['crimes']
        ...                             .obs_to_edge, graph=False)
        >>> counts[(50, 165)]
        4
        
        Create a contiguity based W object.
        
        >>> w = ntw.contiguityweights(graph=False)
        
        Using the W object, access to ESDA functionality is provided.
        First, a vector of attributes is created for all edges
        with observations.
        
        >>> w = ntw.contiguityweights(graph=False)
        >>> edges = w.neighbors.keys()
        >>> y = np.zeros(len(edges))
        >>> for i, e in enumerate(edges):
        ...     if e in counts.keys():
        ...         y[i] = counts[e]
        >>> y[3]
        3.0
        
        Next, a standard call ot Moran is made and the
        result placed into `res`.
        
        >>> res = pysal.explore.esda.moran.Moran(y, w, permutations=99)
        >>> type(res)
        <class 'pysal.explore.esda.moran.Moran'>
        
        """
        
        neighbors = {}
        neighbors = OrderedDict()
        
        if graph:
            edges = self.graphedges
        else:
            edges = self.edges
            
        if weightings:
            _weights = {}
        else:
            _weights = None
            
        for key in edges:
            neighbors[key] = []
            if weightings:
                _weights[key] = []
                
            for neigh in edges:
                if key == neigh:
                    continue
                if key[0] == neigh[0] or key[0] == neigh[1]\
                        or key[1] == neigh[0] or key[1] == neigh[1]:
                    neighbors[key].append(neigh)
                    if weightings:
                        _weights[key].append(weightings[neigh])
                # TODO: Add a break condition - everything is sorted,
                #       so we know when we have stepped beyond
                #       a possible neighbor.
                # if key[1] > neigh[1]:  #NOT THIS
                    # break
        w = weights.W(neighbors, weights=_weights)
        
        return w


    def distancebandweights(self, threshold, n_proccess=None, gen_tree=False):
        """Create distance based weights.
        
        Parameters
        ----------
        
        threshold : float
            Distance threshold value.
        
        n_processes : int, str
            (Optional) Specify the number of cores to utilize. Default is 1
            core. Use (int) to specify an exact number or cores. Use ("all")
            to request all available cores.
        
        gen_tree : bool
            rebuild shortest path {True}, or skip {False}
        
        Returns
        -------
        
        w : pysal.lib.weights.weights.W
            A PySAL W Object representing the binary adjacency of the network.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> streets_file = examples.get_path('streets.shp')
        >>> ntw = spgh.Network(in_data=streets_file)
        >>> w = ntw.distancebandweights(threshold=500)
        >>> w.n
        230
        >>> w.histogram
        [(1, 22), (2, 58), (3, 63), (4, 40), (5, 36), (6, 3), (7, 5), (8, 3)]
        
        """
        try:
            hasattr(self.alldistances)
        except AttributeError:
            self.node_distance_matrix(n_proccess, gen_tree=gen_tree)
            
        neighbor_query = np.where(self.distancematrix < threshold)
        neighbors = defaultdict(list)
        for i, n in enumerate(neighbor_query[0]):
            neigh = neighbor_query[1][i]
            if n != neigh:
                neighbors[n].append(neighbor_query[1][i])
        w = weights.W(neighbors)
        
        return w


    def snapobservations(self, in_data, name,
                         idvariable=None, attribute=None):
        """Snap a point pattern shapefile to this network object. The point
        pattern is stored in the network.pointpattern['key'] attribute of
        the network object.
        
        Parameters
        ----------
        
        in_data : geopandas.GeoDataFrame or str
            The input geographic data. Either (1) a path to a shapefile (str);
            or (2) a geopandas.GeoDataFrame.
        
        name : str
            Name to be assigned to the point dataset.
        
        idvariable : str
            Column name to be used as ID variable.
        
        attribute : bool
            Defines whether attributes should be extracted. True for attribute
            extraction. False for no attribute extraaction.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> streets_file = examples.get_path('streets.shp')
        >>> ntw = spgh.Network(in_data=streets_file)
        >>> pt_str = 'crimes'
        >>> in_data = examples.get_path('{}.shp'.format(pt_str))
        >>> ntw.snapobservations(in_data, pt_str, attribute=True)
        >>> ntw.pointpatterns[pt_str].npoints
        287
        
        """
        
        self.pointpatterns[name] = PointPattern(in_data=in_data,
                                                idvariable=idvariable,
                                                attribute=attribute)
        self._snap_to_edge(self.pointpatterns[name])


    def compute_distance_to_nodes(self, x, y, edge):
        """Given an observation on a network edge, return the distance to
        the two nodes that bound that end.
        
        Parameters
        ----------
        
        x : float
            x-coordinate of the snapped point.
        
        y : float
            y-coordiante of the snapped point.
        
        edge : tuple
            (node0, node1) representation of the network edge.
        
        Returns
        -------
        
        d1 : float
            The distance to node0. Always the node with the lesser id.
        
        d2 : float
            The distance to node1. Always the node with the greater id.
        
        """
        
        d1 = util.compute_length((x, y), self.node_coords[edge[0]])
        d2 = util.compute_length((x, y), self.node_coords[edge[1]])
        
        return d1, d2

    def compute_snap_dist(self, pattern, idx):
        """Given an observation snapped to a network edge, calculate the
        distance from the original location to the snapped location.
        
        Parameters
        -----------
        
        pattern : spaghetti.network.PointPattern
            point pattern object
        idx : int
            point id
        
        Return
        ------
        dist : float
            euclidean distance from original location to snapped location.
        """
        loc = pattern.points[idx]['coordinates']
        snp = pattern.snapped_coordinates[idx]
        dist = util.compute_length(loc, snp)
        return dist

    def _snap_to_edge(self, pointpattern):
        """ Used internally to snap point observations to network edges.
        
        Parameters
        -----------
        
        pointpattern : spaghetti.network.PointPattern
            point pattern object
        
        Returns
        -------
        
        obs_to_edge : dict
            Dict with edges as keys and lists of points as values.
        
        edge_to_obs : dict
            Dict with point ids as keys and edge tuples as values.
        
        dist_to_node : dict
            Dict with point ids as keys and values as dicts with keys for node
            ids and values as distances from point to node.
        
        """
        
        obs_to_edge = {}
        dist_to_node = {}
        dist_snapped = {}
        obs_to_node = defaultdict(list)
        
        pointpattern.snapped_coordinates = {}
        segments = []
        s2e = {}
        for edge in self.edges:
            head = self.node_coords[edge[0]]
            tail = self.node_coords[edge[1]]
            segments.append(cg.Chain([head, tail]))
            s2e[(head, tail)] = edge
        
        # snap points
        points = {}
        p2id = {}
        for pointIdx, point in pointpattern.points.items():
            points[pointIdx] = point['coordinates']
        snapped = util.snap_points_on_segments(points, segments)
        
        # record obs_to_edge, dist_to_node, and dist_snapped
        for pointIdx, snapInfo in snapped.items():
            x, y = snapInfo[1].tolist()
            edge = s2e[tuple(snapInfo[0])]
            if edge not in obs_to_edge:
                obs_to_edge[edge] = {}
            obs_to_edge[edge][pointIdx] = (x, y)
            pointpattern.snapped_coordinates[pointIdx] = (x, y)
            d1, d2 = self.compute_distance_to_nodes(x, y, edge)
            dist_to_node[pointIdx] = {edge[0]: d1, edge[1]: d2}
            dist_snapped[pointIdx] = self.compute_snap_dist(pointpattern,
                                                            pointIdx)
        
        # record obs_to_node
        obs_to_node = defaultdict(list)
        for k, v in obs_to_edge.items():
            keys = v.keys()
            obs_to_node[k[0]] = keys
            obs_to_node[k[1]] = keys
        
        pointpattern.obs_to_edge = obs_to_edge
        pointpattern.dist_to_node = dist_to_node
        pointpattern.dist_snapped = dist_snapped
        pointpattern.obs_to_node = list(obs_to_node)


    def count_per_edge(self, obs_on_network, graph=True):
        """Compute the counts per edge.
        
        Parameters
        ----------
        
        obs_on_network : dict
            Dict of observations on the network. {(edge):{pt_id:(coords)}} or
            {edge:[(coord),(coord),(coord)]}
        
        Returns
        -------
        counts : dict
            {(edge):count}
        
        Examples
        --------
        
        Note that this passes the obs_to_edge attribute of a
        point pattern snapped to the network.
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        >>> ntw.snapobservations(examples.get_path('crimes.shp'),
        ...                                           'crimes',
        ...                                           attribute=True)
        
        >>> counts = ntw.count_per_edge(ntw.pointpatterns['crimes']
        ...                             .obs_to_edge, graph=False)
        >>> counts[(140, 142)]
        10
        
        >>> s = sum([v for v in list(counts.values())])
        >>> s
        287
        
        """
        counts = {}
        if graph:
            for key, observations in obs_on_network.items():
                cnt = len(observations)
                if key in self.graph_to_edges.keys():
                    key = self.graph_to_edges[key]
                try:
                    counts[key] += cnt
                except KeyError:
                    counts[key] = cnt
        else:
            for key in obs_on_network.keys():
                counts[key] = len(obs_on_network[key])
        return counts


    def _newpoint_coords(self, edge, distance):
        """ Used internally to compute new point coordinates during snapping.
        """
        x1 = self.node_coords[edge[0]][0]
        y1 = self.node_coords[edge[0]][1]
        x2 = self.node_coords[edge[1]][0]
        y2 = self.node_coords[edge[1]][1]
        if x1 == x2:  # Vertical line case
            x0 = x1
            if y1 < y2:
                y0 = y1 + distance
            elif y1 > y2:
                y0 = y2 + distance
            else:    # Zero length edge
                y0 = y1
            return x0, y0
        m = (y2 - y1) / (x2 - x1)
        if x1 > x2:
            x0 = x1 - distance / np.sqrt(1 + m**2)
        elif x1 < x2:
            x0 = x1 + distance / np.sqrt(1 + m**2)
        y0 = m * (x0 - x1) + y1
        
        return x0, y0


    def simulate_observations(self, count, distribution='uniform'):
        """ Generate a simulated point pattern on the network.
        
        Parameters
        ----------
        
        count : int
            The number of points to create or mean of the distribution
            if not 'uniform'.
        
        distribution : str
            {'uniform', 'poisson'} distribution of random points. If poisson
            if the distrubtion the poisson is calculated from half the total
            network length
        
        Returns
        -------
        
        random_pts : dict
            Keys are the edge tuple. Values are lists of new point coordinates.
        
        Examples
        --------
       
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        >>> ntw.snapobservations(examples.get_path('crimes.shp'),
        ...                                        'crimes',
        ...                                         attribute=True)
       
        >>> npts = ntw.pointpatterns['crimes'].npoints
        >>> sim = ntw.simulate_observations(npts)
        >>> isinstance(sim, spgh.network.SimulatedPointPattern)
        True
        
        """
        simpts = SimulatedPointPattern()
        
        # Cumulative Network Length.
        edges = []
        lengths = np.zeros(len(self.edge_lengths))
        for i, key in enumerate(self.edge_lengths.keys()):
            edges.append(key)
            lengths[i] = self.edge_lengths[key]
        stops = np.cumsum(lengths)
        totallength = stops[-1]
        
        if distribution is 'uniform':
            nrandompts = np.random.uniform(0, totallength, size=(count,))
        elif distribution is 'poisson':
            mid_length = totallength / 2.
            nrandompts = np.random.poisson(mid_length, size=(count,))
            
        for i, r in enumerate(nrandompts):
            idx = np.where(r < stops)[0][0]
            assignment_edge = edges[idx]
            distance_from_start = stops[idx] - r
            # Populate the coordinates dict.
            x0, y0 = self._newpoint_coords(assignment_edge,
                                           distance_from_start)
            simpts.snapped_coordinates[i] = (x0, y0)
            simpts.obs_to_node[assignment_edge[0]].append(i)
            simpts.obs_to_node[assignment_edge[1]].append(i)
            
            # Populate the distance to node.
            distance_from_end = self.edge_lengths[edges[idx]]\
                                                        - distance_from_start
            simpts.dist_to_node[i] = {assignment_edge[0]: distance_from_start,
                                      assignment_edge[1]: distance_from_end}
            
            simpts.points = simpts.snapped_coordinates
            simpts.npoints = len(simpts.points)
            
        return simpts


    def enum_links_node(self, v0):
        """Returns the edges (links) around node.
        
        Parameters
        -----------
        
        v0 : int
            Node id
        
        Returns
        -------
        
        links : list
            List of tuple edges adjacent to the node.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        >>> ntw.enum_links_node(24)
        [(24, 48), (24, 25), (24, 26)]
        
        """
        links = []
        neighbornodes = self.adjacencylist[v0]
        for n in neighbornodes:
            links.append(tuple(sorted([n, v0])))
        
        return links


    def node_distance_matrix(self, n_processes, gen_tree=False):
        """ Called from within allneighbordistances(),
        nearestneighbordistances(), and distancebandweights().
        
        Parameters
        -----------
        
        n_processes : int
            cpu cores for multiprocessing.
       
        gen_tree : bool
            rebuild shortest path {True}, or skip {False}
        
        """
        self.alldistances = {}
        nnodes = len(self.node_list)
        self.distancematrix = np.empty((nnodes, nnodes))
        # Single-core processing
        if not n_processes:
            for node in self.node_list:
                distance, pred = util.dijkstra(self, self.edge_lengths, node)
                pred = np.array(pred)
                if gen_tree:
                    tree = util.generatetree(pred)
                else:
                    tree = None
                self.alldistances[node] = (distance, tree)
                self.distancematrix[node] = distance
        # Multiprocessing
        if n_processes:
            import multiprocessing as mp
            from itertools import repeat
            if n_processes == "all":
                cores = mp.cpu_count()
            else:
                cores = n_processes
            p = mp.Pool(processes=cores)
            distance_pred = p.map(util.dijkstra_multi,
                                  zip(repeat(self),
                                      repeat(self.edge_lengths),
                                      self.node_list))
            iterations = range(len(distance_pred))
            distance = [distance_pred[itr][0] for itr in iterations]
            pred = np.array([distance_pred[itr][1] for itr in iterations])
            for node in self.node_list:
                if gen_tree:
                    tree = util.generatetree(pred[node])
                else:
                    tree = None
                self.alldistances[node] = (distance[node], tree)
                self.distancematrix[node] = distance[node]


    def allneighbordistances(self, sourcepattern, destpattern=None,
                             fill_diagonal=None, n_processes=None,
                             gen_tree=False, snap_dist=False):
        """ Compute either all distances between i and j in a single point
        pattern or all distances between each i from a source pattern and all
        j from a destination pattern.
        
        Parameters
        ----------
        
        sourcepattern : str or spaghetti.network.PointPattern
            The key of a point pattern snapped to the network OR
            the full spaghetti.network.PointPattern object.
        
        destpattern : str
            (Optional) The key of a point pattern snapped to the network OR
            the full spaghetti.network.PointPattern object.
        
        fill_diagonal : float, int
            (Optional) Fill the diagonal of the cost matrix. Default in None
            and will populate the diagonal with numpy.nan Do not declare a
            destpattern for a custom fill_diagonal.
        
        n_processes : int, str
            (Optional) Specify the number of cores to utilize. Default is 1
            core. Use (int) to specify an exact number or cores.  Use ("all")
            to request all available cores.
        
        gen_tree : bool
            rebuild shortest path {True}, or skip {False}
        
        snap_dist : bool
            include the distance from the original location to the snapped
            location along the network. Default is False.
        
        Returns
        -------
        
        nearest : numpy.ndarray
            An array of shape (n,n) storing distances between all points.
        
        tree_nearest : dict
            nearest network node to point pattern node shortest path lookup
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        >>> ntw.snapobservations(examples.get_path('crimes.shp'),
        ...                                        'crimes',
        ...                                         attribute=True)
        
        
        >>> s2s_dist = ntw.allneighbordistances('crimes')
        >>> s2s_dist[0,0], s2s_dist[1,0]
        (nan, 3105.189475447081)
        
        
        >>> ntw.snapobservations(examples.get_path('schools.shp'),
        ...                                        'schools',
        ...                                        attribute=False)
        
        
        >>> s2d_dist = ntw.allneighbordistances('crimes',
        ...                                     destpattern='schools')
        >>> s2d_dist[0,0], s2d_dist[1,0]
        (4520.72353741989, 6340.422971967316)
        
        
        >>> s2d_dist, tree = ntw.allneighbordistances('schools',
        ...                                           gen_tree=True)
        >>> tree[(6, 7)]
        (173, 64)
        """
        
        if not hasattr(self, 'alldistances'):
            self.node_distance_matrix(n_processes, gen_tree=gen_tree)
        
        if type(sourcepattern) is str:
            sourcepattern = self.pointpatterns[sourcepattern]
            if destpattern:
                destpattern = self.pointpatterns[destpattern]
        
        # Source setup
        src_indices = list(sourcepattern.points.keys())
        src_d2n = copy.deepcopy(sourcepattern.dist_to_node)
        nsource_pts = len(src_indices)
        src_nodes = {}
        for s in src_indices:
            e1, e2 = src_d2n[s].keys()
            src_nodes[s] = (e1, e2)
            
        # Destination setup
        symmetric = False
        if destpattern is None:
            symmetric = True
            destpattern = sourcepattern
        dest_indices = list(destpattern.points.keys())
        dst_d2n = copy.deepcopy(destpattern.dist_to_node)
        ndest_pts = len(dest_indices)
        dest_searchpts = copy.deepcopy(dest_indices)
        dest_nodes = {}
        
        # add snapping distance to each pointpattern
        if snap_dist:
            patterns = [sourcepattern, destpattern]
            dist_copies = [src_d2n, dst_d2n]
            for elm, pp in enumerate(patterns):
                for pidx, dists_dict in dist_copies[elm].items():
                    for nidx, ndist in dists_dict.items():
                        dists_dict[nidx] = ndist + pp.dist_snapped[pidx]
        
        for s in dest_indices:
            e1, e2 = dst_d2n[s].keys()
            dest_nodes[s] = (e1, e2)
            
        # Output setup
        nearest = np.empty((nsource_pts, ndest_pts))
        nearest[:] = np.inf
        tree_nearest = {}
        
        for p1 in src_indices:
            # Get the source nodes and dist to source nodes.
            source1, source2 = src_nodes[p1]
            set1 = set(src_nodes[p1])
            # Distance from node1 to p, distance from node2 to p.
            sdist1, sdist2 = src_d2n[p1].values()
            
            if symmetric:
                # Only compute the upper triangle if symmetric.
                dest_searchpts.remove(p1)
            for p2 in dest_searchpts:
                dest1, dest2 = dest_nodes[p2]
                set2 = set(dest_nodes[p2])
                if set1 == set2:  # same edge
                    x1, y1 = sourcepattern.snapped_coordinates[p1]
                    x2, y2 = destpattern.snapped_coordinates[p2]
                    computed_length = util.compute_length((x1, y1), (x2, y2))
                    nearest[p1, p2] = computed_length
                    
                else:
                    ddist1, ddist2 = dst_d2n[p2].values()
                    
                    d11 = self.distancematrix[source1][dest1]
                    d21 = self.distancematrix[source2][dest1]
                    d12 = self.distancematrix[source1][dest2]
                    d22 = self.distancematrix[source2][dest2]
                    
                    # Find the shortest distance from the path passing
                    # through each of the two origin nodes to the first
                    # destination node.
                    sd_1 = d11 + sdist1
                    sd_21 = d21 + sdist2
                    sp_combo1 = source1, dest1
                    if sd_1 > sd_21:
                        sd_1 = sd_21
                        sp_combo1 = source2, dest1
                    
                    # Now add the point to node one distance on
                    # the destination edge.
                    len_1 = sd_1 + ddist1
                    
                    # Repeat the prior but now for the paths entering
                    # at the second node of the second edge.
                    sd_2 = d12 + sdist1
                    sd_22 = d22 + sdist2
                    sp_combo2 = source1, dest2
                    if sd_2 > sd_22:
                        sd_2 = sd_22
                        sp_combo2 = source2, dest2
                    len_2 = sd_2 + ddist2
                    
                    # Now find the shortest distance path between point
                    # 1 on edge 1 and point 2 on edge 2, and assign.
                    sp_12 = len_1
                    s_node, d_node = sp_combo1
                    if len_1 > len_2:
                        sp_12 = len_2
                        s_node, d_node = sp_combo2
                    
                    nearest[p1, p2] = sp_12
                    tree_nearest[p1, p2] = (s_node, d_node)
                    
                if symmetric:
                    # Mirror the upper and lower triangle
                    # when symmetric.
                    nearest[p2, p1] = nearest[p1, p2]
                    
        # Populate the main diagonal when symmetric.
        if symmetric:
            if fill_diagonal is None:
                np.fill_diagonal(nearest, np.nan)
            else:
                np.fill_diagonal(nearest, fill_diagonal)
        
        if gen_tree:
            return nearest, tree_nearest
        else:
            return nearest


    def nearestneighbordistances(self, sourcepattern, destpattern=None,
                                 n_processes=None, gen_tree=False,
                                 all_dists=None, snap_dist=False,
                                 keep_zero_dist=True):
        """Compute the interpattern nearest neighbor distances or the
        intrapattern nearest neighbor distances between a source
        pattern and a destination pattern.
        
        Parameters
        ----------
        
        sourcepattern : str
            The key of a point pattern snapped to the network.
        
        destpattern : str
            (Optional) The key of a point pattern snapped to the network.
        
        n_processes : int, str
            (Optional) Specify the number of cores to utilize. Default is 1
            core. Use (int) to specify an exact number or cores. Use ("all")
            to request all available cores.
        
        gen_tree : bool
            rebuild shortest path {True}, or skip {False}
        
        all_dists : numpy.ndarray
            An array of shape (n,n) storing distances between all points.
        
        snap_dist : bool
            include the distance from the original location to the snapped
            location along the network. Default is False.
        
        keep_zero_dist : bool
            Include zero values in minimum distance (True) or exclude (False).
            Default is True. If the source pattern is the same as the
            destination pattern the diagonal is filled with nans
        
        Returns
        -------
        nearest : dict
            key is source point id, value is tuple of list containing 
            nearest destination point ids and distance.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        >>> ntw.snapobservations(examples.get_path('crimes.shp'), 'crimes')
        >>> nn = ntw.nearestneighbordistances('crimes', keep_zero_dist=True)
        >>> nn[11], nn[18]
        (([18, 19], 165.33982412719126), ([19], 0.0))
        
        >>> nn = ntw.nearestneighbordistances('crimes', keep_zero_dist=False)
        >>> nn[11], nn[18]
        (([18, 19], 165.33982412719126), ([11], 165.33982412719126))
        
        """
        if sourcepattern not in self.pointpatterns.keys():
            err_msg = "Available point patterns are {}"
            raise KeyError(err_msg.format(self.pointpatterns.keys()))
            
        if not hasattr(self, 'alldistances'):
            self.node_distance_matrix(n_processes, gen_tree=gen_tree)
        
        symmetric = sourcepattern != destpattern
        
        # (for source-to-source patterns) if zero-distance neighbors are
        # desired, keep the diagonal as NaN and take the minimum distance
        # neighbor(s), which may include zero distance neighors.
        fill_diagonal = None
        if not keep_zero_dist and symmetric:
            # (for source-to-source patterns) if zero-distance neighbors should
            # be ignored, convert the diagonal to 0.0 and take the minimum
            # distance neighbor(s) that is/are not 0.0 distance.
            fill_diagonal = 0.
        
        sourcepattern = self.pointpatterns[sourcepattern]
        if destpattern:
            destpattern = self.pointpatterns[destpattern]
        
        if all_dists is None:
            all_dists = self.allneighbordistances(sourcepattern,
                                                  destpattern=destpattern,
                                                  fill_diagonal=fill_diagonal,
                                                  n_processes=n_processes,
                                                  gen_tree=gen_tree,
                                                  snap_dist=snap_dist)
        nearest = {}
        
        for source_index in sourcepattern.points.keys():
            if keep_zero_dist and symmetric:
                val = np.nanmin(all_dists[source_index,:])
            else:
                val = np.min(all_dists[source_index,:]\
                                      [np.nonzero(all_dists[source_index,:])])
            # nearest destination (may be more than one if equal distance)
            dest_idxs = np.where(all_dists[source_index,:] == val)[0].tolist()
            nearest[source_index] = (dest_idxs, val)
            
        return nearest


    def NetworkF(self, pointpattern, nsteps=10, permutations=99, threshold=0.2,
                 distribution='uniform',  lowerbound=None, upperbound=None):
        """Computes a network constrained F-Function
        
        Parameters
        ----------
        
        pointpattern : spaghetti.network.PointPattern
            A spaghetti point pattern object.
        
        nsteps : int
            The number of steps at which the count of the nearest neighbors
            is computed.
        
        permutations : int
            The number of permutations to perform (default 99).
        
        threshold : float
            The level at which significance is computed.
            -- 0.5 would be 97.5% and 2.5%
        
        distribution : str
            The distribution from which random points are sampled
            -- uniform or poisson
        
        lowerbound : float
            The lower bound at which the F-function is computed. (Default 0).
        
        upperbound : float
            The upper bound at which the F-function is computed. Defaults to
            the maximum observed nearest neighbor distance.
        
        Returns
        -------
        
        NetworkF : spaghetti.analysis.NetworkF
            A network F class instance.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(in_data=examples.get_path('streets.shp'))
        >>> pt_str = 'crimes'
        >>> in_data = examples.get_path('{}.shp'.format(pt_str))
        >>> ntw.snapobservations(in_data, pt_str, attribute=True)
        >>> crimes = ntw.pointpatterns['crimes']
        >>> sim = ntw.simulate_observations(crimes.npoints)
        >>> fres = ntw.NetworkF(crimes, permutations=5, nsteps=10)
        >>> fres.lowerenvelope.shape[0]
        10
        """
        return NetworkF(self, pointpattern, nsteps=nsteps,
                        permutations=permutations, threshold=threshold,
                        distribution=distribution, lowerbound=lowerbound,
                        upperbound=upperbound)


    def NetworkG(self, pointpattern, nsteps=10, permutations=99,
                 threshold=0.5, distribution='uniform',
                 lowerbound=None, upperbound=None):
        """Computes a network constrained G-Function
        
        Parameters
        ----------
        
        pointpattern : spaghetti.network.PointPattern
            A spaghetti point pattern object.
        
        nsteps : int
            The number of steps at which the count of the nearest neighbors
            is computed.
        
        permutations : int
            The number of permutations to perform (default 99).
        
        threshold : float
            The level at which significance is computed.
            -- 0.5 would be 97.5% and 2.5%
        
        distribution : str
            The distribution from which random points are sampled
            -- uniform or poisson
        
        lowerbound : float
            The lower bound at which the G-function is computed. (Default 0).
        
        upperbound : float
            The upper bound at which the G-function is computed. Defaults to
            the maximum observed nearest neighbor distance.
        
        Returns
        -------
        
        NetworkG : spaghetti.analysis.NetworkG
            A network G class instance.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(in_data=examples.get_path('streets.shp'))
        >>> pt_str = 'crimes'
        >>> in_data = examples.get_path('{}.shp'.format(pt_str))
        >>> ntw.snapobservations(in_data, pt_str, attribute=True)
        >>> crimes = ntw.pointpatterns['crimes']
        >>> sim = ntw.simulate_observations(crimes.npoints)
        >>> gres = ntw.NetworkG(crimes, permutations=5, nsteps=10)
        >>> gres.lowerenvelope.shape[0]
        10
        """
        
        return NetworkG(self, pointpattern, nsteps=nsteps,
                        permutations=permutations, threshold=threshold,
                        distribution=distribution, lowerbound=lowerbound,
                        upperbound=upperbound)


    def NetworkK(self, pointpattern, nsteps=10, permutations=99,
                 threshold=0.5, distribution='uniform',
                 lowerbound=None, upperbound=None):
        """
        Computes a network constrained K-Function
        
        Parameters
        ----------
        
        pointpattern : spaghetti.network.PointPattern
            A spaghetti point pattern object.
        
        nsteps : int
            The number of steps at which the count of the nearest neighbors
            is computed.
        
        permutations : int
            The number of permutations to perform (default 99).
        
        threshold : float
            The level at which significance is computed.
            -- 0.5 would be 97.5% and 2.5%
        
        distribution : str
            The distribution from which random points are sampled
            -- uniform or poisson
        
        lowerbound : float
            The lower bound at which the K-function is computed. (Default 0).
        
        upperbound : float
            The upper bound at which the K-function is computed. Defaults to
            the maximum observed nearest neighbor distance.
        
        Returns
        -------
        
        NetworkK : spaghetti.analysis.NetworkK
            A network K class instance.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(in_data=examples.get_path('streets.shp'))
        >>> pt_str = 'crimes'
        >>> in_data = examples.get_path('{}.shp'.format(pt_str))
        >>> ntw.snapobservations(in_data, pt_str, attribute=True)
        >>> crimes = ntw.pointpatterns['crimes']
        >>> sim = ntw.simulate_observations(crimes.npoints)
        >>> kres = ntw.NetworkK(crimes, permutations=5, nsteps=10)
        >>> kres.lowerenvelope.shape[0]
        10
        """
        return NetworkK(self, pointpattern, nsteps=nsteps,
                        permutations=permutations, threshold=threshold,
                        distribution=distribution, lowerbound=lowerbound,
                        upperbound=upperbound)


    def segment_edges(self, distance):
        """Segment all of the edges in the network at either a
        fixed distance or a fixed number of segments.
        
        Parameters
        -----------
        
        distance : float
            The distance at which edges are split.
        
        Returns
        -------
        
        sn : spaghetti.Network
            spaghetti Network object.
        
       Examples
        --------
       
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        >>> n200 = ntw.segment_edges(200.0)
        >>> len(n200.edges)
        688
        
        """
        
        sn = Network()
        sn.adjacencylist = copy.deepcopy(self.adjacencylist)
        sn.edge_lengths = copy.deepcopy(self.edge_lengths)
        sn.edges = set(copy.deepcopy(self.edges))
        sn.node_coords = copy.deepcopy(self.node_coords)
        sn.node_list = copy.deepcopy(self.node_list)
        sn.nodes = copy.deepcopy(self.nodes)
        sn.pointpatterns = copy.deepcopy(self.pointpatterns)
        sn.in_data = self.in_data
        
        current_node_id = max(self.nodes.values())
        
        newedges = set()
        removeedges = set()
        for e in sn.edges:
            length = sn.edge_lengths[e]
            interval = distance
            
            totallength = 0
            currentstart = startnode = e[0]
            endnode = e[1]
            
            # If the edge will be segmented remove the current
            # edge from the adjacency list.
            if interval < length:
                sn.adjacencylist[e[0]].remove(e[1])
                sn.adjacencylist[e[1]].remove(e[0])
                sn.edge_lengths.pop(e, None)
                removeedges.add(e)
            else:
                continue
                
            while totallength < length:
                currentstop = current_node_id
                if totallength + interval > length:
                    currentstop = endnode
                    interval = length - totallength
                    totallength = length
                else:
                    current_node_id += 1
                    currentstop = current_node_id
                    totallength += interval
                    
                    # Compute the new node coordinate.
                    newx, newy = self._newpoint_coords(e, totallength)
                    
                    # Update node_list.
                    if currentstop not in sn.node_list:
                        sn.node_list.append(currentstop)
                    
                    # Update nodes and node_coords.
                    sn.node_coords[currentstop] = newx, newy
                    sn.nodes[(newx, newy)] = currentstop
                
                # Update the adjacency list.
                sn.adjacencylist[currentstart].append(currentstop)
                sn.adjacencylist[currentstop].append(currentstart)
                
                # Add the new edge to the edge dict.
                # Iterating over this so we need to add after iterating.
                newedges.add(tuple(sorted([currentstart, currentstop])))
                
                # Modify edge_lengths.
                current_start_stop = tuple(sorted([currentstart, currentstop]))
                sn.edge_lengths[current_start_stop] = interval
                
                # Increment the start to the stop.
                currentstart = currentstop
        
        sn.edges.update(newedges)
        sn.edges.difference_update(removeedges)
        sn.edges = list(sn.edges)
        # Update the point pattern snapping.
        for instance in sn.pointpatterns.values():
            sn._snap_to_edge(instance)
            
        return sn


    def savenetwork(self, filename):
        """Save a network to disk as a binary file.
        
        Parameters
        ----------
        
        filename : str
            The filename where the network should be saved. This should be a
            full path or it will be save in the current directory.
        
        Examples
        --------
        
        >>> import pysal.explore.spaghetti as spgh
        >>> ntw = spgh.Network(examples.get_path('streets.shp'))
        >>> ntw.savenetwork('mynetwork.pkl')
        """
        with open(filename, 'wb') as networkout:
            pickle.dump(self, networkout, protocol=2)

    @staticmethod
    def loadnetwork(filename):
        """Load a network from a binary file saved on disk.
        
        Parameters
        ----------
        
        filename : str
            The filename where the network should be saved.
        
        Returns
        -------
        
        self : spaghetti.Network
            spaghetti Network object
            
        """
        with open(filename, 'rb') as networkin:
            self = pickle.load(networkin)
            
        return self


class PointPattern():
    """A stub point pattern class used to store a point pattern. This class is
    monkey patched with network specific attributes when the points are snapped
    to a network. In the future this class may be replaced with a generic point
    pattern class.
    
    Parameters
    ----------
    
    in_data : geopandas.GeoDataFrame or str
        The input geographic data. Either (1) a path to a shapefile (str);
        or (2) a geopandas.GeoDataFrame.
    
    idvariable : str
        Field in the shapefile to use as an id variable.
    
    attribute :  bool
        {False, True} A flag to indicate whether all attributes are tagged
        to this class.
    
    Attributes
    ----------
    
    points : dict
        Keys are the point ids (int). Values are the x,y coordinates (tuple).
    
    npoints : int
        The number of points.
    
    obs_to_edge : dict
        Keys are edge ids (tuple). Values are snapped point information (dict).
        Withing the snapped point information (dict) keys are observation id
        (int), and values are snapped coordinates.
    
    obs_to_node : list
       list of incident network nodes to snapped observation points converted
       from default_dict. Originally in the form of paired left/right nearest
       network nodes - {netnode1: obs_id1, netnode2: obs_id1,
       netnode1: obs_id2... netnode1: obs_idn}, then simplified to a list in
       the form - [netnode1, netnode2, netnode1, netnode2, ...].
       
    dist_to_node : dict
        Keys are observations ids (int). Values are distance lookup (dict).
        Within distance lookup (dict) keys are the two incident nodes of the
        edge and values are distance to each of those edges.
    
    snapped_coordinates : dict
        Keys are the point ids (int). Values are the snapped x,y
        coordinates (tuple).
    
    snap_dist : bool
        include the distance from the original location to the snapped
        location along the network. Default is False.
    
    """
    def __init__(self, in_data=None, idvariable=None, attribute=False):
        self.points = {}
        self.npoints = 0
        
        if isinstance(in_data, str):
            from_shp = True
        else:
            from_shp = False
            
        if idvariable:
            ids = weights.util.get_ids(in_data, idvariable)
        else:
            ids = None
            
        if from_shp:
            pts = open(in_data)
        else:
            pts_objs = list(in_data.geometry)
            pts = [cg.shapes.Point((p.x, p.y)) for p in pts_objs]
            
        # Get attributes if requested
        if attribute:
            if from_shp:
                dbname = os.path.splitext(in_data)[0] + '.dbf'
                db = open(dbname)
            else:
                db = in_data.drop(in_data.geometry.name,
                                   axis=1).values.tolist()
                db = [[d] for d in db]
        else:
            db = None
            
        for i, pt in enumerate(pts):
            # ids, attributes
            if ids and db is not None:
                self.points[ids[i]] = {'coordinates': pt, 'properties': db[i]}
            # ids, no attributes
            elif ids and db is None:
                self.points[ids[i]] = {'coordinates': pt, 'properties': None}
            # no ids, attributes
            elif not ids and db is not None:
                self.points[i] = {'coordinates': pt, 'properties': db[i]}
            # no ids, no attributes
            else:
                self.points[i] = {'coordinates': pt, 'properties': None}
        if from_shp:
            pts.close()
            if db:
                db.close()
                
        self.npoints = len(self.points.keys())


class SimulatedPointPattern():
    """Struct style class to mirror the Point Pattern Class. If the
    PointPattern class has methods, it might make sense to make this a child of
    that class. This class is not intended to be used by the external user.
    
    Attributes
    ----------
    
    npoints : int
        The number of points.
    
    obs_to_edge : dict
        Keys are edge ids (tuple). Values are snapped point information (dict).
        Withing the snapped point information (dict) keys are observation id
        (int), and values are snapped coordinates.
    
    obs_to_node : list
       list of incident network nodes to snapped observation points converted
       from default_dict. Originally in the form of paired left/right nearest
       network nodes - {netnode1: obs_id1, netnode2: obs_id1,
       netnode1: obs_id2... netnode1: obs_idn}, then simplified to a list in
       the form - [netnode1, netnode2, netnode1, netnode2, ...].
       
    dist_to_node : dict
        Keys are observations ids (int). Values are distance lookup (dict).
        Within distance lookup (dict) keys are the two incident nodes of the
        edge and values are distance to each of those edges.
    
    snapped_coordinates : dict
        Keys are the point ids (int). Values are the snapped x,y
        coordinates (tuple).
    """
    def __init__(self):
        self.npoints = 0
        self.obs_to_edge = {}
        self.obs_to_node = defaultdict(list)
        self.dist_to_node = {}
        self.snapped_coordinates = {}


class SortedEdges(OrderedDict):
    """
    Parameters
    ----------
    OrderedDict : collections.OrderedDict
    
    """
    def next_key(self, key):
        """
        Parameters
        ----------
        
        key : 
            
        Returns
        -------
        
        n : 
            
        """
        next = self._OrderedDict__map[key][1]
        if next is self._OrderedDict__root:
            raise ValueError("{!r} is the last key.".format(key))
        n = next[2]
        return n

    def first_key(self):
        """
        
        Returns
        -------
        
        key : 
            
        """
        for key in self:
            return key
        raise ValueError("No sorted edges remain.")



from __future__ import division

import numpy as np
import scipy.stats as stats
import copy

from numpy import linalg as la
from warnings import warn as Warn
from ...abstracts import Sampler_Mixin, Hashmap, Trace
from ... import verify
from ...utils import ind_covariance, chol_mvn


SAMPLERS = ['Alphas', 'Betas', 'Sigma2', 'Tau2']

class Base_MVCM(Sampler_Mixin):
    """
    The class that actually ends up setting up the MVCM model. Sets configs,
    data, truncation, and starting parameters, and then attempts to apply the
    sample function n_samples times to the state.
    """
    def __init__(self, Y, X, Delta,
                 n_samples=1000, n_jobs=1,
                 extra_traced_params = None,
                 priors=None,
                 starting_values=None):
        super(Base_MVCM, self).__init__()

        N, p = X.shape
        _N, J = Delta.shape
        self.state = Hashmap(**{'X':X, 'Y':Y, 'Delta':Delta,
                           'N':N, 'J':J, 'p':p })
        self.traced_params = copy.deepcopy(SAMPLERS)
        extras = extra_traced_params
        if extras is not None:
            self.traced_params.extend(extra_tracked_params)
        hashmaps = [Hashmap(**{k:[] for k in self.traced_params})]*n_jobs
        self.trace = Trace(*hashmaps)

        if priors is None:
            priors = dict()
        if starting_values is None:
            starting_values = dict()

        self._setup_priors(**priors)
        self._setup_starting_values(**starting_values)

        self.cycles = 0
        self.configs = None


        try:
            self.sample(n_samples, n_jobs=n_jobs)
        except (np.linalg.LinAlgError, ValueError) as e:
            Warn('Encountered the following LinAlgError. '
                 'Model will return for debugging. \n {}'.format(e))


    def _setup_priors(self, Sigma2_a0=.001, Sigma2_b0 = .001,
                      Betas_cov0 = None, Betas_mean0 = None,
                      Tau2_a0 = .001, Tau2_b0 = .001):
        ## Prior specifications
        st = self.state
        st.Sigma2_a0 =  Sigma2_a0
        st.Sigma2_b0 = Sigma2_b0
        if Betas_cov0 is None:
            Betas_cov0 = np.eye(self.state.p) * 100
        st.Betas_cov0 = Betas_cov0
        if Betas_mean0 is None:
            Betas_mean0 = np.zeros((self.state.p, 1))
        st.Betas_mean0 = Betas_mean0
        st.Tau2_a0 = Tau2_a0
        st.Tau2_b0 = Tau2_b0

    def _finalize(self):
        """
        This computes derived properties of hyperparameters that do not change
        over iterations. This is called one time before sampling.
        """
        st = self.state

        st.XtX = np.dot(st.X.T, st.X)
        st.DeltatDelta = np.dot(st.Delta.T, st.Delta)

        st.Betas_cov0i = np.linalg.inv(st.Betas_cov0)
        st.Betas_covm = np.dot(st.Betas_cov0, st.Betas_mean0)
        st.Sigma2_an = st.N / 2 + st.Sigma2_a0
        st.Tau2_an = st.J / 2 + st.Tau2_a0
        st.In = np.identity(st.N)
        st.Ij = np.identity(st.J)
        st.DeltaAlphas = np.dot(st.Delta, st.Alphas)
        st.XBetas = np.dot(st.X, st.Betas)

        st.Psi_1 = ind_covariance
        st.Psi_2 = ind_covariance

        st.PsiRho = st.In
        st.PsiLambda = st.Ij

        st.PsiSigma2 = st.In * st.Sigma2
        st.PsiSigma2i = la.inv(st.PsiSigma2)
        st.PsiTau2 = st.Ij * st.Tau2
        st.PsiTau2i = la.inv(st.PsiTau2)

        st.PsiRhoi = la.inv(st.PsiRho)
        st.PsiLambdai = la.inv(st.PsiLambda)

    def _setup_starting_values(self, Betas=None, Alphas=None,
                                Sigma2=4, Tau2=4):
        """
        Set abrbitrary starting values for the Metropolis sampler
        """
        if Betas is None:
            Betas = np.zeros((self.state.p, 1))
        if Alphas is None:
            Alphas = np.zeros((self.state.J, 1))
        self.state.Betas = Betas
        self.state.Alphas = Alphas
        self.state.Sigma2 = Sigma2
        self.state.Tau2 = Tau2

    def _iteration(self):
        st = self.state

        ### Sample the Beta conditional posterior
        ### P(beta | . ) \propto L(Y|.) \dot P(\beta)
        ### is
        ### N(Sb, S) where
        ### S = (X' Sigma^{-1}_Y X + S_0^{-1})^{-1}
        ### b = X' Sigma^{-1}_Y (Y - Delta Alphas) + S^{-1}\mu_0
        covm_update = st.X.T.dot(st.X) / st.Sigma2
        covm_update += st.Betas_cov0i
        covm_update = la.inv(covm_update)

        resids = st.Y - st.Delta.dot(st.Alphas)
        XtSresids = st.X.T.dot(resids) / st.Sigma2
        mean_update = XtSresids + st.Betas_cov0i.dot(st.Betas_mean0)
        mean_update = np.dot(covm_update, mean_update)
        st.Betas = chol_mvn(mean_update, covm_update)
        st.XBetas = np.dot(st.X, st.Betas)

        ### Sample the Random Effect conditional posterior
        ### P( Alpha | . ) \propto L(Y|.) \dot P(Alpha | \lambda, Tau2)
        ###                               \dot P(Tau2) \dot P(\lambda)
        ### is
        ### N(Sb, S)
        ### Where
        ### S = (Delta'Sigma_Y^{-1}Delta + Sigma_Alpha^{-1})^{-1}
        ### b = (Delta'Sigma_Y^{-1}(Y - X\beta) + 0)
        covm_update = st.Delta.T.dot(st.Delta) / st.Sigma2
        covm_update += st.PsiTau2i
        covm_update = la.inv(covm_update)

        resids = st.Y - st.XBetas
        mean_update = st.Delta.T.dot(resids) / st.Sigma2
        mean_update = np.dot(covm_update, mean_update)
        st.Alphas = chol_mvn(mean_update, covm_update)
        st.DeltaAlphas = np.dot(st.Delta, st.Alphas)

        ### Sample the Random Effect aspatial variance parameter
        ### P(Tau2 | .) \propto L(Y|.) \dot P(\Alpha | \lambda, Tau2)
        ###                            \dot P(Tau2) \dot P(\lambda)
        ### is
        ### IG(J/2 + a0, u'(\Psi(\lambda))^{-1}u * .5 + b0)
        bn = st.Alphas.T.dot(st.PsiLambdai).dot(st.Alphas) * .5 + st.Tau2_b0
        st.Tau2 = stats.invgamma.rvs(st.Tau2_an, scale=bn)
        st.PsiTau2 = st.Ij * st.Tau2
        st.PsiTau2i = la.inv(st.PsiTau2)

        ### Sample the response aspatial variance parameter
        ### P(Sigma2 | . ) \propto L(Y | .) \dot P(Sigma2)
        ### is
        ### IG(N/2 + a0, eta'Psi(\rho)^{-1}eta * .5 + b0)
        ### Where eta is the linear predictor, Y - X\beta + \DeltaAlphas
        eta = st.Y - st.XBetas - st.DeltaAlphas
        bn = eta.T.dot(st.PsiRhoi).dot(eta) * .5 + st.Sigma2_b0
        st.Sigma2 = stats.invgamma.rvs(st.Sigma2_an, scale=bn)

class MVCM(Base_MVCM):
    """
    This is a class that provides the generic structures required for the two-level variance components model

    Formally, this is the following distributional model for Y:

    Y ~ N(X * Beta, I_n * Sigma2 + Delta Delta' Tau2)

    Where Delta is the dummy variable matrix, Sigma2 is the response-level variance, and Tau2 is the region-level variance. 

    Arguments
    ----------
    Y       :   numpy.ndarray
                The (n,1) array containing the response to be modeled
    X       :   numpy.ndarray
                The (n,p) array containing covariates used to predict the  response, Y. Can be None, and if so,
                 will be used to fit a naive model with only a response-level constant.
    Z       :   numpy.ndarray
                The (J,p') array of region-level covariates used to predict the response, Y.
    Delta   :   numpy.ndarray
                The (n,J) dummy variable matrix, relating observation i to region j. Required if membership is not passed.
    membership: numpy.ndarray
                The (n,) vector of labels assigning each observation to a region. Required if Delta is not passed.
    transform  : string
                 weights transform to use in modeling.
    verbose    : bool
                 whether to provide verbose output about sampling progress
    n_samples : int
                the number of samples to draw. If zero, the model will only be initialized, and later sampling can be done using model.sample
    n_jobs  :   int
                the number of parallel chains to run. Defaults to 1.
    extra_traced_param  :   list of strings
                            extra parameters in the trace to record.
    center  :   bool
                Whether to center the input data so that its mean is zero. Occasionally improves the performance of the sampler.
    scale   :   bool
                Whether to rescale the input data so that its variance is one. May dramatically improve the performance of the sampler, at the cost of interpretability of the coefficients.
    priors  :   dictionary
                A dictionary used to configure the priors for the model. This may include the following keys:
                    Betas_cov0  : prior covariance for Beta parameter vector
                                (default: I*100)
                    Betas_mean0 : prior mean for Beta parameters
                                (default: 0)
                    Sigma2_a0   : prior shape parameter for inverse gamma prior on response-level variance
                                (default: .001)
                    Sigma2_b0   : prior scale parameter for inverse gamma prior on response-level variance
                                (default: .001)
                    Tau2_a0     : prior shape parameter for inverse gamma prior on regional-level variance
                                (default: .001)
                    Tau2_b0     : prior scale parameter for inverse gamma prior on regional-level variance
                                (default: .001)

    starting_value :    dictionary
                        A dictionary containing the starting values of the sampler. If n_jobs > 1, these starting values are perturbed randomly to induce overdispersion.

                        This dicutionary may include the following keys:
                        Betas   : starting vector of Beta parameters.
                                (default: np.zeros(p,1))
                        Alphas  : starting vector of Alpha variance components.
                                (default: np.zeros(J,1))
                        Sigma2  : starting value of response-level variance
                                (default: 4)
                        Tau2    : starting value of region-level varaince
                                (default: 4)

                For options that can be in Lambda/Rho_configs, see:
                pysal.model.spvcm.steps.Slice, pysal.model.spvcm.steps.Metropolis
    """
    def __init__(self, Y, X, Z=None, Delta=None, membership=None,
                 #data options
                 transform ='r', n_samples=1000, n_jobs=1,
                 verbose=False,
                 extra_traced_params = None,
                 priors=None,
                 starting_values=None,
                 center=False,
                 scale=False
                 ):
        if X is None:
            X = np.ones_like(Y)
            center=False
            scale=False
        N, _ = X.shape
        if Delta is not None:
            _,J = Delta.shape
        elif membership is not None:
            J = len(np.unique(membership))
        else:
            raise UserWarning("No Delta matrix nor membership classification provided. Refusing to arbitrarily assign units to upper-level regions.")
        Delta, membership = verify.Delta_members(Delta, membership, N, J)

        if Z is not None:
            Z = Delta.dot(Z)
            X = np.hstack((X,Z))
        if center:
            X = verify.center(X)
        if scale:
            X = verify.scale(X)

        X = verify.covariates(X)

        self._verbose = verbose
        super(MVCM, self).__init__(Y, X, Delta, n_samples=n_samples,
                                   n_jobs=n_jobs,
                                   extra_traced_params = extra_traced_params,
                                   priors=priors,
                                   starting_values=starting_values)



import numpy as np
import pysal.viz.mapclassify as classify
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.cm as cm

"""
Utility functions for lightweight visualizations in splot
"""

__author__ = ("Stefanie Lumnitz <stefanie.lumitz@gmail.com>")


def moran_hot_cold_spots(moran_loc, p=0.05):
    sig = 1 * (moran_loc.p_sim < p)
    HH = 1 * (sig * moran_loc.q == 1)
    LL = 3 * (sig * moran_loc.q == 3)
    LH = 2 * (sig * moran_loc.q == 2)
    HL = 4 * (sig * moran_loc.q == 4)
    cluster = HH + LL + LH + HL
    return cluster


def mask_local_auto(moran_loc, p=0.5):
    '''
    Create Mask for coloration and labeling of local spatial autocorrelation

    Parameters
    ----------
    moran_loc : esda.moran.Moran_Local instance
        values of Moran's I Global Autocorrelation Statistic
    p : float
        The p-value threshold for significance. Points will
        be colored by significance.

    Returns
    -------
    cluster_labels : list of str
        List of labels - ['ns', 'HH', 'LH', 'LL', 'HL']
    colors5 : list of str
        List of colours - ['lightgrey', 'red', 'lightblue','blue', 'pink']
    colors : array of str
        Array containing coloration for each input value/ shape.
    labels : list of str
        List of label for each attribute value/ polygon.
    '''
    # create a mask for local spatial autocorrelation
    cluster = moran_hot_cold_spots(moran_loc, p)

    cluster_labels = ['ns', 'HH', 'LH', 'LL', 'HL']
    labels = [cluster_labels[i] for i in cluster]

    colors5 = {0: 'lightgrey',
               1: '#d7191c',
               2: '#abd9e9',
               3: '#2c7bb6',
               4: '#fdae61'}
    colors = [colors5[i] for i in cluster]  # for Bokeh
    # for MPL:
    colors5 = (['#d7191c', '#fdae61', '#abd9e9', '#2c7bb6', 'lightgrey'])

    # HACK need this, because MPL sorts these labels while Bokeh does not
    cluster_labels.sort()
    return cluster_labels, colors5, colors, labels


_classifiers = {
    'box_plot': classify.Box_Plot,
    'equal_interval': classify.Equal_Interval,
    'fisher_jenks': classify.Fisher_Jenks,
    'headtail_breaks': classify.HeadTail_Breaks,
    'jenks_caspall': classify.Jenks_Caspall,
    'jenks_caspall_forced': classify.Jenks_Caspall_Forced,
    'max_p_classifier': classify.Max_P_Classifier,
    'maximum_breaks': classify.Maximum_Breaks,
    'natural_breaks': classify.Natural_Breaks,
    'quantiles': classify.Quantiles,
    'percentiles': classify.Percentiles,
    'std_mean': classify.Std_Mean,
    'user_defined': classify.User_Defined,
    }


def bin_values_choropleth(attribute_values, method='quantiles',
                          k=5):
    '''
    Create bins based on different classification methods.
    Needed for legend labels and Choropleth coloring.

    Parameters
    ----------
    attribute_values : array or geopandas.series instance
        Array containing relevant attribute values.
    method : str
        Classification method to be used. Options supported:
        * 'quantiles' (default)
        * 'fisher-jenks'
        * 'equal-interval'
    k : int
        Number of bins, assigning values to. Default k=5

    Returns
    -------
    bin_values : mapclassify instance
        Object containing bin ids for each observation (.yb),
        upper bounds of each class (.bins), number of classes (.k)
        and number of onservations falling in each class (.counts)
    '''
    if method not in ['quantiles', 'fisher_jenks', 'equal_interval']:
        raise ValueError("Method {} not supported".format(method))

    bin_values = _classifiers[method](attribute_values, k)
    return bin_values


def bin_labels_choropleth(gdf, attribute_values, method='quantiles', k=5):
    '''
    Create labels for each bin in the legend

    Parameters
    ----------
    gdf : Geopandas dataframe
        Dataframe containign relevant shapes and attribute values.
    attribute_values : array or geopandas.series instance
        Array containing relevant attribute values.
    method : str, optional
        Classification method to be used. Options supported:
        * 'quantiles' (default)
        * 'fisher-jenks'
        * 'equal-interval'
    k : int, optional
        Number of bins, assigning values to. Default k=5

    Returns
    -------
    bin_labels : list of str
        List of label for each bin.
    '''
    # Retrieve bin values from bin_values_choropleth()
    bin_values = bin_values_choropleth(attribute_values, method=method, k=k)

    # Extract bin ids (.yb) and upper bounds for each class (.bins)
    yb = bin_values.yb
    bins = bin_values.bins

    # Create bin labels (smaller version)
    bin_edges = bins.tolist()
    bin_labels = []
    for i in range(k):
        bin_labels.append('<{:1.1f}'.format(bin_edges[i]))

    # Add labels (which are the labels printed in the legend) to each row of gdf
    labels = np.array([bin_labels[c] for c in yb])
    gdf['labels_choro'] = [str(l) for l in labels]
    return bin_labels


def add_legend(fig, labels, colors):
    """
    Add a legend to a figure given legend labels & colors.

    Parameters
    ----------
    fig : Bokeh Figure instance
        Figure instance labels should be generated for.
    labels : list of str
        Labels to use as legend entries.
    colors : Bokeh Palette instance
        Palette instance containing colours of choice.
    """
    from bokeh.models import Legend
    # add labels to figure (workaround,
    # legend with geojsondatasource doesn't work,
    # see https://github.com/bokeh/bokeh/issues/5904)
    items = []
    for label, color in zip(labels, colors):
        patch = fig.patches(xs=[], ys=[], fill_color=color)
        items.append((label, [patch]))

    legend = Legend(items=items, location='top_left', margin=0,
                    orientation='horizontal')
    # possibility to define glyph_width=10, glyph_height=10)
    legend.label_text_font_size = '8pt'
    fig.add_layout(legend, 'below')
    return legend


def format_legend(values):
    """
    Helper to return sensible legend values
    
    Parameters
    ----------
    values: array
        Values plotted in legend.
    """
    in_thousand = False
    if np.any(values > 1000):
        in_thousand = True
        values = values / 1000
    return values, in_thousand


def calc_data_aspect(plot_height, plot_width, bounds):
    # Deal with data ranges in Bokeh:
    # make a meter in x and a meter in y the same in pixel lengths
    aspect_box = plot_height / plot_width   # 2 / 1 = 2
    xmin, ymin, xmax, ymax = bounds
    x_range = xmax - xmin  # 1 = 1 - 0
    y_range = ymax - ymin  # 3 = 3 - 0
    aspect_data = y_range / x_range  # 3 / 1 = 3
    if aspect_data > aspect_box:
        # we need to increase x_range,
        # such that aspect_data becomes equal to aspect_box
        halfrange = 0.5 * x_range * (aspect_data / aspect_box - 1)
        # 0.5 * 1 * (3 / 2 - 1) = 0.25
        xmin -= halfrange  # 0 - 0.25 = -0.25
        xmax += halfrange  # 1 + 0.25 = 1.25
    else:
        # we need to increase y_range
        halfrange = 0.5 * y_range * (aspect_box / aspect_data - 1)
        ymin -= halfrange
        ymax += halfrange

    # Add a bit of margin to both x and y
    margin = 0.03
    xmin -= (xmax - xmin) / 2 * margin
    xmax += (xmax - xmin) / 2 * margin
    ymin -= (ymax - ymin) / 2 * margin
    ymax += (ymax - ymin) / 2 * margin
    return xmin, xmax, ymin, ymax


# Utility functions for colormaps
# Color design
splot_colors = dict(moran_base='#bababa',
                    moran_fit='#d6604d')

# Utility function #1 - forces continuous diverging colormap to be centered at zero
def shift_colormap(cmap, start=0, midpoint=0.5, stop=1.0, name='shiftedcmap'):
    '''
    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and you want the
    middle of the colormap's dynamic range to be at zero

    Parameters
    ----------
    cmap : matplotlib colormap
        colormap to be altered
    start : float, optional
        Offset from lowest point in the colormap's range.
        Should be between 0.0 and `midpoint`.
        Default =0.0 (no lower ofset).
    midpoint : float, optional
        The new center of the colormap.Should be between 0.0 and
        1.0. In general, this should be 1 - vmax/(vmax + abs(vmin)).
        For example if your data range from -15.0 to +5.0 and
        you want the center of the colormap at 0.0, `midpoint`
        should be set to  1 - 5/(5 + 15)) or 0.75.
        Default =0.5 (no shift).
    stop : float, optional
        Offset from highets point in the colormap's range.
        Should be between `midpoint` and 1.0.
        Default =1.0 (no upper ofset).
    name : str, optional
        Name of the new colormap.
    
    Returns
    -------
    new_cmap : A new colormap that has been shifted. 
    '''
    if isinstance(cmap, str):
        cmap = cm.get_cmap(cmap) 

    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(start, stop, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False), 
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    new_cmap = mpl.colors.LinearSegmentedColormap(name, cdict)
    plt.register_cmap(cmap=new_cmap)
    return new_cmap


# Utility #2 - truncate colorcap in order to grab only positive or negative portion
def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=100):
    '''
    Function to truncate a colormap by selecting a subset of
    the original colormap's values

    Parameters
    ----------
    cmap : Mmatplotlib colormap
        Colormap to be altered
    minval : float, optional
        Minimum value of the original colormap to include
        in the truncated colormap. Default =0.0.
    maxval : Maximum value of the original colormap to
        include in the truncated colormap. Default =1.0.
    n : int, optional
        Number of intervals between the min and max values
        for the gradient of the truncated colormap. Default =100.
          
    Returns
    -------
    new_cmap : A new colormap that has been shifted. 
    '''
    new_cmap = mpl.colors.LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(np.linspace(minval, maxval, n)))
    return new_cmap
